import pytest
import requests

from pactman import Pact, Term


def test_regex_passes():
    pact = Pact('Consumer', 'Provider', file_write_mode='never').given('spam'). \
        with_request('GET', '/', headers={'Spam': Term(r'\w+', 'spam')}).will_respond_with(200)
    with pact:
        requests.get(pact.uri, headers={'Spam': 'ham'})


def test_regex_fails():
    pact = Pact('Consumer', 'Provider', file_write_mode='never').given('spam'). \
        with_request('GET', '/', headers={'Spam': Term(r'\w+', 'spam')}).will_respond_with(200)
    with pact:
        with pytest.raises(AssertionError):
            requests.get(pact.uri, headers={'Spam': '!@#$'})


def test_regex_passes_v3():
    pact = Pact('Consumer', 'Provider', file_write_mode='never', version='3.0.0').given('spam'). \
        with_request('GET', '/', headers={'Spam': Term(r'\w+', 'spam')}).will_respond_with(200)
    with pact:
        requests.get(pact.uri, headers={'Spam': 'ham'})


def test_regex_fails_v3():
    pact = Pact('Consumer', 'Provider', file_write_mode='never', version='3.0.0').given('spam'). \
        with_request('GET', '/', headers={'Spam': Term(r'\w+', 'spam')}).will_respond_with(200)
    with pact:
        with pytest.raises(AssertionError):
            requests.get(pact.uri, headers={'Spam': '!@#$'})



import json
import logging
import sys
from unittest.mock import Mock

import semver

from pactman.test.test_verifier import FakeResponse
from pactman.verifier.result import LoggedResult
from pactman.verifier.verify import ResponseVerifier


logging.basicConfig(level=logging.DEBUG)

result = LoggedResult()
version = sys.argv[1]
verifier = ResponseVerifier
response = FakeResponse

with open(sys.argv[2]) as file:
    case = json.load(file)

    pact = Mock(provider='SpamProvider', consumer='SpamConsumer', version=version,
                semver=semver.parse(version))
    verifier(pact, case['expected'], result).verify(response(case['actual']))



import pytest

from pactman import EachLike, Like, Term


def test_default_options():
    assert EachLike(1).ruby_protocol() == {'json_class': 'Pact::ArrayLike', 'contents': 1, 'min': 1}


def test_minimum():
    assert EachLike(1, minimum=2).ruby_protocol() == {'json_class': 'Pact::ArrayLike', 'contents': 1, 'min': 2}


def test_minimum_assertion_error():
    with pytest.raises(AssertionError) as e:
        EachLike(1, minimum=0)
    assert str(e.value) == 'Minimum must be greater than or equal to 1'


def test_nested_matchers():
    matcher = EachLike({'username': Term('[a-z]+', 'user'), 'id': Like(123)})
    assert matcher.ruby_protocol() == {
        'json_class': 'Pact::ArrayLike',
        'contents': {
            'username': {
                'json_class': 'Pact::Term',
                'data': {
                    'matcher': {
                        'json_class': 'Regexp',
                        's': '[a-z]+',
                        'o': 0},
                    'generate': 'user'}},
            'id': {
                'json_class': 'Pact::SomethingLike',
                'contents': 123}},
        'min': 1
    }



from pactman import Consumer, EachLike, Provider
from pactman.mock.request import Request
from pactman.mock.response import Response


def test_eachlike():
    pact = (
        Consumer('consumer').has_pact_with(Provider('provider'), version='2.0.0')
        .given("the condition exists")
        .upon_receiving("a request")
        .with_request("POST", "/path", body=EachLike(1))
        .will_respond_with(200, body={"results": EachLike(1)})
    )

    result = pact.construct_pact(pact._interactions[0])
    assert result == {
        'consumer': {'name': 'consumer'},
        'provider': {'name': 'provider'},
        'interactions': [
            {
                'description': 'a request',
                'providerState': 'the condition exists',
                'request': dict(method='POST', path='/path', body=[1],
                                matchingRules={'$.body': {'match': 'type', 'min': 1}}),
                'response': dict(status=200, body={'results': [1]},
                                 matchingRules={'$.body.results': {'match': 'type', 'min': 1}}),

            }
        ],
        'metadata': dict(pactSpecification=dict(version='2.0.0'))
    }


def test_falsey_request_body():
    target = Request('GET', '/path', body=[])
    assert target.json('2.0.0') == {'method': 'GET', 'path': '/path', 'body': []}


def test_falsey_response_body():
    target = Response(200, body=[])
    assert target.json('2.0.0') == {'status': 200, 'body': []}



from functools import total_ordering


@total_ordering
class Part:
    def __init__(self, value, params):
        self.value = value
        self.params = params

    def has_param(self, name):
        for k, v in self.params:
            if k == name:
                return True
        return False

    def __repr__(self):
        return f'<Part {", ".join(self.value + ["%s=%s"%i for i in self.params])}>'

    def __eq__(self, other):
        return (self.value, self.params) == (other.value, other.params)

    def __lt__(self, other):
        return (self.value, self.params) < (other.value, other.params)


def _parseparam(s, marker):
    while s[:1] == marker:
        s = s[1:]
        end = s.find(marker)
        while end > 0 and (s.count('"', 0, end) - s.count('\\"', 0, end)) % 2:
            end = s.find(marker, end + 1)
        if end < 0:
            end = len(s)
        f = s[:end]
        yield f.strip()
        s = s[end:]


def parse_header(line):
    parts = _parseparam(';' + line, ';')
    for part in parts:
        params = []
        key = []
        for option in _parseparam(',' + part, ','):
            i = option.find('=')
            if i >= 0:
                name = option[:i].strip().lower()
                value = option[i+1:].strip()
                if len(value) >= 2 and value[0] == value[-1] == '"':
                    value = value[1:-1]
                    value = value.replace('\\\\', '\\').replace('\\"', '"')
                params.append((name, value))
            else:
                key.append(option)
        yield Part(key, params)


def get_header_param(header, name):
    for header_part in parse_header(header):
        for param, value in header_part.params:
            if param == name:
                return value



# -*- encoding: utf8 -*-
from unittest.mock import Mock, call, patch

import urllib3
import urllib3.poolmanager
from urllib3.response import HTTPResponse

from pactman.mock.mock_urlopen import MockURLOpenHandler, patcher


def test_patched_urlopen_calls_service_with_request_parameters():
    pact = Mock(port=1234)
    mock_service = Mock(pact=pact, return_value=HTTPResponse())
    try:
        patcher.add_service(mock_service)
        http = urllib3.PoolManager()
        response = http.request('GET', 'http://api.test:1234/path')
    finally:
        patcher.remove_service(mock_service)
    assert mock_service.call_args == call('GET', '/path', body=None, headers={})
    assert response is mock_service.return_value


@patch.object(urllib3.connectionpool.HTTPConnectionPool, 'urlopen')
def test_patched_urlopen_handles_many_positional_arguments(HTTPConnectionPool_urlopen):
    # urllib3 passes in up to 7 positional arguments to urlopen so we need to ensure
    # our mocked urlopen method handles these
    mock_service = Mock(config=Mock(port=1234), return_value=HTTPResponse())
    try:
        patcher.add_service(mock_service)
        pool = urllib3.poolmanager.pool_classes_by_scheme['http']('api.test', port=5678)
        pool.urlopen('POST', '/path', 'body1', {}, None, True, False)
    finally:
        patcher.remove_service(mock_service)
    expected_call = call('POST', '/path', 'body1', {}, None, True, False)
    assert HTTPConnectionPool_urlopen.call_args == expected_call


def test_urlopen_responder_handles_json_body():
    h = MockURLOpenHandler(Mock())

    interaction = dict(
        response=dict(body={'message': 'hello world'}, status=200)
    )
    r = h.respond_for_interaction(interaction)

    assert r.data == b'{"message": "hello world"}'
    assert r.headers['Content-Type'] == 'application/json; charset=UTF-8'


def test_urlopen_responder_handles_json_string_body():
    h = MockURLOpenHandler(Mock())

    interaction = dict(
        response=dict(body="hello world", status=200)
    )
    r = h.respond_for_interaction(interaction)

    assert r.data == b'"hello world"'
    assert r.headers['Content-Type'] == 'application/json; charset=UTF-8'


def test_urlopen_responder_handles_json_encoding():
    h = MockURLOpenHandler(Mock())

    interaction = dict(
        response=dict(
            headers={'content-type': 'application/json; charset=utf-8'},
            body="héllo world", status=200,
        ),
    )
    r = h.respond_for_interaction(interaction)

    assert r.data == b'"h\\u00e9llo world"'
    assert r.headers['Content-Type'] == 'application/json; charset=utf-8'


def test_urlopen_responder_handles_non_json_body():
    h = MockURLOpenHandler(Mock())

    interaction = dict(
        response=dict(
            headers={'content-type': 'text/plain; charset=utf-8'},
            body="héllo world".encode('utf-8'), status=200,
        ),
    )
    r = h.respond_for_interaction(interaction)

    assert r.data == b'h\xc3\xa9llo world'
    assert r.headers['Content-Type'] == 'text/plain; charset=utf-8'



import json
import pathlib
from itertools import chain
from unittest.mock import Mock

import requests

import pytest
import semver
from pactman.verifier.broker_pact import BrokerPact, BrokerPacts, pact_id
from pactman.verifier.result import Result
from pactman.verifier.verify import (Interaction, RequestVerifier,
                                     ResponseVerifier)
from restnavigator import Navigator

BASE_DIR = pathlib.Path(__file__).absolute().parents[0]


def all_testcases(path, version):
    for entry in path.iterdir():
        if entry.is_file() and entry.suffix == '.json':
            yield str(entry), version
        elif entry.is_dir():
            yield from all_testcases(entry, version)


@pytest.fixture
def fake_interaction():
    return {
        'description': 'dummy',
        'request': {'method': 'GET', 'path': '/users-service/user/alex'},
        'response': {'headers': {}, 'status': 200}
    }


@pytest.fixture
def fake_pact(fake_interaction):
    return {
        'provider': {'name': 'SpamProvider'},
        'consumer': {'name': 'SpamConsumer'},
        'interactions': [fake_interaction],
        'metadata': {'pact-specification': {'version': '2.0.0'}},
    }


@pytest.fixture
def mock_pact():
    def create_mock(version):
        return Mock(provider='SpamProvider', consumer='SpamConsumer', version=version,
                    semver=semver.parse(version))
    return create_mock


@pytest.fixture
def mock_result():
    result = Result()
    result.fail = Mock(return_value=False)
    result.warn = Mock()
    return result


@pytest.fixture
def mock_result_factory(mock_result):
    return Mock(return_value=mock_result)


def test_pact_id():
    assert pact_id(1) == repr(1)


def test_pact_loading(monkeypatch, fake_pact):
    p = BrokerPacts('SpamProvider', 'http://broker.example/')
    mock_pact = Mock(fetch=Mock(return_value=fake_pact))
    mock_provider = Mock(fetch=lambda: None, __getitem__=lambda s, k: [mock_pact])
    monkeypatch.setattr(Navigator, 'hal', lambda url, default_curie=None: {
        'latest-provider-pacts': lambda provider=None: mock_provider
    })

    for e in p.consumers():
        assert str(e) == 'Pact between consumer SpamConsumer and provider SpamProvider'
        assert e.provider == 'SpamProvider'

    for e in p.all_interactions():
        assert e.description == 'dummy'


def test_pact_publish_uses_interaction_result(fake_pact):
    result_factory = Mock(return_value=Mock(success=True))
    mock_result_publisher = Mock()
    broker_pact = {'publish-verification-results': mock_result_publisher}
    p = BrokerPact(fake_pact, result_factory, broker_pact)
    p.publish_result('1.0')
    mock_result_publisher.create.assert_called_once()
    payload = mock_result_publisher.create.call_args[0][0]
    assert payload == {"success": True, "providerApplicationVersion": "1.0"}


@pytest.mark.parametrize('first_success, second_success, combined_success', [
    (False, True, False),
    (True, False, False),
    (True, True, True),
])
def test_pact_publish_aggregates_interaction_results(
    fake_pact, fake_interaction, first_success, second_success, combined_success
):
    result_factory = Mock()
    fake_pact = {**fake_pact, 'interactions': [fake_interaction, fake_interaction]}
    result_factory.side_effect = [Mock(success=first_success), Mock(success=second_success)]
    mock_result_publisher = Mock()
    broker_pact = {'publish-verification-results': mock_result_publisher}
    BrokerPact(fake_pact, result_factory, broker_pact).publish_result('1.0')
    payload = mock_result_publisher.create.call_args[0][0]
    assert payload == {"success": combined_success, "providerApplicationVersion": "1.0"}


def test_pact_file_loading(fake_interaction):
    p = BrokerPact.load_file(str(BASE_DIR / 'testcases-version-3' / 'dummypact.json'))
    assert str(p) == 'Pact between consumer SpamConsumer and provider SpamProvider'
    assert p.provider == 'SpamProvider'
    for e in p.interactions:
        assert e.description == 'dummy'


def test_interaction(mock_pact, mock_result_factory, fake_interaction):
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    assert repr(i) == '<Interaction SpamConsumer:dummy>'


def test_interaction_verify_get(monkeypatch, mock_pact, mock_result_factory, fake_interaction):
    monkeypatch.setattr(requests, 'post', Mock())
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')

    requests.post.assert_not_called()
    requests.get.assert_called_with('http://provider.example/users-service/user/alex', headers={})
    i.response.verify.assert_called()


def test_interaction_verify_method_not_supported(monkeypatch, mock_pact, mock_result_factory, fake_interaction):
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'FLEBBLE'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    i.result.fail.assert_called_with("Request method FLEBBLE not implemented in verifier")
    i.response.verify.assert_not_called()


def test_interaction_verify_qs(monkeypatch, mock_pact, mock_result_factory, fake_interaction):
    monkeypatch.setattr(requests, 'post', Mock())
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['query'] = 'a=b&c=d'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')

    requests.get.assert_called_with(
        'http://provider.example/users-service/user/alex',
        params=dict(a=['b'], c=['d']),
        headers={},
    )
    i.response.verify.assert_called()


def test_interaction_verify_post(monkeypatch, mock_pact, mock_result_factory, fake_interaction):
    monkeypatch.setattr(requests, 'post', Mock())
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'POST'
    fake_interaction['request']['body'] = 'spam'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')

    requests.get.assert_not_called()
    requests.post.assert_called_with(
        'http://provider.example/users-service/user/alex',
        json='spam',
        headers={},
    )
    i.response.verify.assert_called()


def test_interaction_verify_post_unsupported_content_type(
    monkeypatch, mock_pact, mock_result_factory, fake_interaction
):
    monkeypatch.setattr(requests, 'post', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'POST'
    fake_interaction['request']['headers'] = {'Content-Type': 'spam/ham'}
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    i.result.fail.assert_called_with("POST content type spam/ham not implemented in verifier")

    requests.post.assert_not_called()


def test_interaction_verify_delete(monkeypatch, mock_pact, mock_result_factory):
    monkeypatch.setattr(requests, 'delete', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction = {
        'description': 'dummy',
        'request': {'method': 'DELETE', 'path': '/diary-notes/diary-note/1'},
        'response': {'headers': {}, 'status': 200}
    }
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')

    requests.delete.assert_called_with('http://provider.example/diary-notes/diary-note/1', headers={})
    i.response.verify.assert_called()


def test_interaction_verify_put(monkeypatch, mock_pact, mock_result_factory, fake_interaction):
    monkeypatch.setattr(requests, 'put', Mock())
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'PUT'
    fake_interaction['request']['body'] = 'spam'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')

    requests.get.assert_not_called()
    requests.put.assert_called_with(
        'http://provider.example/users-service/user/alex',
        json='spam',
        headers={},
    )
    i.response.verify.assert_called()


def test_interaction_verify_put_unsupported_content_type(monkeypatch, mock_pact, mock_result_factory, fake_interaction):
    monkeypatch.setattr(requests, 'put', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'PUT'
    fake_interaction['request']['headers'] = {'Content-Type': 'spam/ham'}
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    i.result.fail.assert_called_with("PUT content type spam/ham not implemented in verifier")

    requests.put.assert_not_called()


def test_interaction_verify_patch(monkeypatch, mock_pact, mock_result_factory, fake_interaction):
    monkeypatch.setattr(requests, 'patch', Mock())
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'PATCH'
    fake_interaction['request']['body'] = 'spam'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    requests.get.assert_not_called()
    requests.patch.assert_called_with(
        'http://provider.example/users-service/user/alex',
        json='spam',
        headers={},
    )
    i.response.verify.assert_called()


def test_interaction_verify_patch_unsupported_content_type(
    monkeypatch, mock_pact, mock_result_factory, fake_interaction
):
    monkeypatch.setattr(requests, 'patch', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = 'PATCH'
    fake_interaction['request']['headers'] = {'Content-Type': 'spam/ham'}
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    i.result.fail.assert_called_with("PATCH content type spam/ham not implemented in verifier")
    requests.patch.assert_not_called()


@pytest.mark.parametrize('method', ['GET', 'POST', 'DELETE', 'PUT', 'PATCH'])
def test_interaction_sends_headers(monkeypatch, mock_pact, mock_result_factory, fake_interaction, method):
    headers = {'key1': 'value1'}
    requests_method = Mock()
    monkeypatch.setattr(requests, method.lower(), requests_method)
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['request']['method'] = method
    fake_interaction['request']['body'] = 'body-data'
    fake_interaction['request']['headers'] = headers
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    request_kwargs = requests_method.call_args[1]
    assert request_kwargs["headers"] == headers


@pytest.mark.parametrize('option, arg', [('providerState', 'state'), ('providerStates', 'states')])
def test_interaction_verify_with_setup(monkeypatch, mock_pact, mock_result_factory, fake_interaction, option, arg):
    monkeypatch.setattr(requests, 'post', Mock(return_value=Mock(status_code=200)))
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction[option] = 'some state'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')

    requests.post.assert_called_with(
        'http://provider.example/pact-setup/',
        json={'provider': "SpamProvider", 'consumer': "SpamConsumer", arg: "some state"}
    )
    i.response.verify.assert_called()


@pytest.mark.parametrize('option', ['providerState', 'providerStates'])
def test_interaction_setup_fails(monkeypatch, mock_pact, mock_result_factory, fake_interaction, option):
    monkeypatch.setattr(requests, 'post', Mock(return_value=Mock(status_code=400, text='fail')))
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction[option] = 'some state'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/')
    i.result.warn.assert_called_with("Invalid provider state 'some state'")


def test_interaction_setup_connection_fails(monkeypatch, mock_pact, mock_result_factory, fake_interaction):
    monkeypatch.setattr(requests, 'post', Mock(side_effect=requests.exceptions.ConnectionError('barf')))
    monkeypatch.setattr(requests, 'get', Mock())
    # monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['providerState'] = 'some state'
    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.set_versioned_provider_state('http://provider.example/pact-setup/', 'state', 'some state')
    i.result.fail.assert_called_once()


def test_response_verifier(fake_interaction, mock_pact):
    fake_interaction['response']['body'] = dict(a='b', c=1)  # note: no rule for matching of c
    fake_interaction['response']['matchingRules'] = {
        '$.body.a': dict(match='type'),
    }
    fake_interaction['response']['headers'] = {'Content-Type': 'json-yeah'}
    r = ResponseVerifier(mock_pact('2.0.0'), fake_interaction['response'], Mock())
    r.verify(Mock(status_code=200, headers={'Content-Type': 'json-yeah'}, json=Mock(return_value=dict(a='b', c='c'))))


@pytest.mark.parametrize("interaction", [
    {'body': {'key1': 1}},
    {'body': {'key1': 1}, 'matchingRules': {'$.body.key1': {'match': 'type'}}},
])
def test_ResponseVerifier_reports_correct_message_when_key_missing(mock_pact, interaction):
    mock_result = Mock()
    r = ResponseVerifier(mock_pact('2.0.0'), interaction, mock_result)
    r.verify(FakeResponse({"body": {}}))
    mock_result.fail.assert_called_with("Response element 'key1' is missing", ['body'])


@pytest.mark.parametrize("interaction", [
    {'body': {'key1': 1}},
    {'body': {'key1': 1}, 'matchingRules': {'$.body.key1': {'match': 'type'}}},
])
def test_RequestVerifier_reports_correct_message_when_key_missing(mock_pact, interaction):
    mock_result = Mock()
    r = RequestVerifier(mock_pact('2.0.0'), interaction, mock_result)
    r.verify(FakeRequest({"body": {}}))
    mock_result.fail.assert_called_with("Request element 'key1' is missing", ['body'])


def test_ResponseVerifier_checks_for_missing_headers(mock_pact):
    interaction = {"headers": {"X-Custom-Header": "value1"}}
    mock_result = Mock()
    r = ResponseVerifier(mock_pact("2.0.0"), interaction, mock_result)
    assert r.verify(FakeResponse({"headers": {}})) is False
    mock_result.fail.assert_called_with("Response missing header 'X-Custom-Header'")


def test_RequestVerifier_checks_for_missing_headers(mock_pact):
    interaction = {"headers": {"X-Custom-Header": "value1"}}
    mock_result = Mock()
    r = RequestVerifier(mock_pact("2.0.0"), interaction, mock_result)
    assert r.verify(FakeRequest({"headers": {}})) is False
    mock_result.fail.assert_called_with("Request missing header 'X-Custom-Header'")


def test_provider_state_uses_custom_headers(monkeypatch, mock_pact, mock_result_factory, fake_interaction):
    fake_interaction['request']['headers'] = dict(request_header='value')
    monkeypatch.setattr(requests, 'post', Mock(return_value=Mock(status_code=200)))
    monkeypatch.setattr(requests, 'get', Mock())
    monkeypatch.setattr(ResponseVerifier, 'verify', Mock())
    fake_interaction['providerStates'] = 'some state'

    i = Interaction(mock_pact('2.0.0'), fake_interaction, mock_result_factory)
    i.verify('http://provider.example/', 'http://provider.example/pact-setup/',
             extra_provider_headers={'some_header': 'some_header'})

    requests.post.assert_called_with(
        'http://provider.example/pact-setup/',
        json={'provider': "SpamProvider", 'consumer': "SpamConsumer", 'states': "some state"},
        headers={'some_header': "some_header"}
    )

    requests.get.assert_called_with('http://provider.example/users-service/user/alex',
                                    headers={'request_header': 'value', 'some_header': "some_header"})
    i.response.verify.assert_called()


class FakeResponse:
    status = 200
    body = None
    text = 'fake response text'

    def __init__(self, attrs):
        self.__dict__.update(attrs)
        self.status_code = self.status

    def __repr__(self):
        return f'<FakeResponse {self.__dict__}>'

    def json(self):
        return self.body


class FakeRequest:
    method = 'GET'
    body = None

    def __init__(self, attrs):
        self.__dict__.update(attrs)

    def __repr__(self):
        return f'<FakeRequest {self.__dict__}>'

    def json(self):
        return self.body


@pytest.mark.parametrize(
    'filename, version, verifier, response',
    chain(
        ((f, v, ResponseVerifier, FakeResponse) for f, v in chain(
            all_testcases(BASE_DIR / 'pact-spec-version-1.1' / 'testcases' / 'response', '1.1.0'),
            all_testcases(BASE_DIR / 'pact-spec-version-2' / 'testcases' / 'response', '2.0.0'),
            all_testcases(BASE_DIR / 'pact-spec-version-3' / 'testcases' / 'response', '3.0.0'),
            all_testcases(BASE_DIR / 'testcases-version-2' / 'response', '2.0.0'),
            all_testcases(BASE_DIR / 'testcases-version-3' / 'response', '3.0.0'),
        )),
        ((f, v, RequestVerifier, FakeRequest) for f, v in chain(
            all_testcases(BASE_DIR / 'pact-spec-version-1.1' / 'testcases' / 'request', '1.1.0'),
            all_testcases(BASE_DIR / 'pact-spec-version-2' / 'testcases' / 'request', '2.0.0'),
            all_testcases(BASE_DIR / 'pact-spec-version-3' / 'testcases' / 'request', '3.0.0'),
            all_testcases(BASE_DIR / 'testcases-version-2' / 'request', '2.0.0'),
            all_testcases(BASE_DIR / 'testcases-version-3' / 'request', '3.0.0'),
        ))
    )
)
def test_testcases(filename, version, verifier, response, mock_pact, mock_result):
    if filename.endswith(' xml.json'):
        # some of the files don't declare the damned content-type!
        raise pytest.skip('XML content type not supported')
    with open(filename) as file:
        try:
            case = json.load(file)
        except json.JSONDecodeError:
            raise pytest.skip('JSON test case mal-formed')
        if case['expected'].get('headers', {}).get('Content-Type', "") == 'application/xml':
            raise pytest.skip('XML content type not supported')
        rv = verifier(mock_pact(version), case['expected'], mock_result)
        result = rv.verify(response(case['actual']))
        success = not bool(rv.result.fail.call_count)
        assert result == success, 'fail() was not called for failure here'
        assert case['match'] == success



import logging
from urllib.parse import parse_qs, urljoin

import requests

from .matching_rule import (RuleFailed, fold_type, nice_type, rule_matchers_v2,
                            rule_matchers_v3)
from .parse_header import parse_header
from .paths import format_path

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class ProviderStateError(Exception):
    pass


class ProviderStateMissing(ProviderStateError):
    pass


class Interaction:
    def __init__(self, pact, interaction, result_factory):
        self.pact = pact
        self.result = result_factory()
        self.description = interaction['description']
        self.request = interaction['request']
        self.providerState = interaction.get('providerState')
        self.providerStates = interaction.get('providerStates')
        self.response = ResponseVerifier(pact, interaction['response'], self.result)
        self.extra_provider_headers = {}

    def __repr__(self):
        return f"<Interaction {self.pact.consumer}:{self.description}>"

    def __str__(self):
        return f"{self.pact.consumer} with request '{self.description}'"

    def verify(self, service_url, setup_url, extra_provider_headers={}):
        self.extra_provider_headers = extra_provider_headers
        self.result.start(self)
        try:
            self.set_provider_state_with_url(setup_url)
            if self.result.success:
                self.result.success = self.run_service(service_url)
        finally:
            self.result.end()

    def verify_with_callable_setup(self, service_url, provider_setup):
        self.result.start(self)
        try:
            try:
                self.set_provider_state(provider_setup)
            except ProviderStateMissing as e:
                self.result.warn(f'Unable to configure provider state: {e}')
            self.result.success = self.run_service(service_url)
        finally:
            self.result.end()

    def run_service(self, service_url):
        method = self.request['method'].upper()
        handler = getattr(self, f'service_{method}', None)
        if handler is None:
            return self.result.fail(f'Request method {method} not implemented in verifier')
        return handler(service_url)

    def service_GET(self, service_url):
        if 'query' in self.request:
            query = self.request['query']
            if isinstance(query, str):
                # version 2 spec used strings, version 3 uses objects
                query = parse_qs(query)
            r = requests.get(self._get_url(service_url), params=query, headers=self._request_headers)
        else:
            r = requests.get(self._get_url(service_url), headers=self._request_headers)
        return self.response.verify(r)

    def service_HEAD(self, service_url):
        if 'query' in self.request:
            query = self.request['query']
            if isinstance(query, str):
                # version 2 spec used strings, version 3 uses objects
                query = parse_qs(query)
            r = requests.head(self._get_url(service_url), params=query, headers=self._request_headers)
        else:
            r = requests.head(self._get_url(service_url), headers=self._request_headers)
        return self.response.verify(r)

    def service_POST(self, service_url):
        if not self._content_type_json:
            return self.result.fail(f'POST content type {self._content_type} not implemented in verifier')
        r = requests.post(self._get_url(service_url), json=self._request_payload, headers=self._request_headers)
        return self.response.verify(r)

    def service_DELETE(self, service_url):
        r = requests.delete(self._get_url(service_url), headers=self._request_headers)
        return self.response.verify(r)

    def service_PUT(self, service_url):
        if not self._content_type_json:
            return self.result.fail(f'PUT content type {self._content_type} not implemented in verifier')
        r = requests.put(self._get_url(service_url), json=self._request_payload, headers=self._request_headers)
        return self.response.verify(r)

    def service_PATCH(self, service_url):
        if not self._content_type_json:
            return self.result.fail(f'PATCH content type {self._content_type} not implemented in verifier')
        r = requests.patch(self._get_url(service_url), json=self._request_payload, headers=self._request_headers)
        return self.response.verify(r)

    def set_provider_state(self, provider_setup):
        if self.providerState is not None:
            log.debug(f'Setting up provider state {self.providerState!r}')
            provider_setup(self.providerState)
            return
        for state in self.providerStates:
            if 'name' not in state:
                raise KeyError(f'provider state missing "name": {state}')
            name = state['name']
            params = state.get('params', {})
            log.debug(f"Setting up provider state {name!r} with params {params}")
            provider_setup(state['name'], **params)
            if params:
                log.info(f'Using provider state {name} with params {params}')
            else:
                log.info(f'Using provider state {name}')

    def set_provider_state_with_url(self, setup_url):
        if self.providerState is not None:
            return self.set_versioned_provider_state(setup_url, 'state', self.providerState)
        elif self.providerStates is not None:
            return self.set_versioned_provider_state(setup_url, 'states', self.providerStates)
        return True

    def set_versioned_provider_state(self, setup_url, var, state):
        log.debug(f'Setting up provider state {state!r}')
        kwargs = dict(json={
            'provider': self.pact.provider,
            'consumer': self.pact.consumer,
            var: state
        })
        if self.extra_provider_headers:
            kwargs['headers'] = self.extra_provider_headers
        try:
            r = requests.post(setup_url, **kwargs)
        except requests.exceptions.ConnectionError as e:
            try:
                # try to pull the actual error out of the nested (nested (nested, string-ified))) exception
                reason = str(e.args[0].reason).split(": ", 1)[1]
            except Exception:
                reason = str(e)
            return self.result.fail(f'Unable to configure provider state {state!r} at {setup_url}: {reason}')
        if r.status_code != 200:
            text = repr(r.text)
            if len(text) > 60:
                text = text[:60] + '...' + text[-1]
            log.debug(f'HTTP {r.status_code} from provider state setup URL: {text}')
            return self.result.warn(f'Invalid provider state {state!r}')
        log.info(f'Using provider state {state!r}')
        return True

    def _get_url(self, service_url):
        return urljoin(service_url, self.request['path'])

    @property
    def _request_headers(self):
        headers = dict(self.extra_provider_headers)
        headers.update(self.request.get('headers', {}))
        return headers

    @property
    def _request_payload(self):
        return self.request.get('body')

    @property
    def _content_type_json(self):
        return 'application/json' in self._content_type

    @property
    def _content_type(self):
        return self._request_headers.get('Content-Type', 'application/json')


def rules_present(item):
    return "present" if item is not MISSING else "absent"


class ResponseVerifier:
    interaction_name = 'Response'

    def __init__(self, pact, interaction, result):
        self.pact = pact
        self.status = interaction.get('status', MISSING)
        self.headers = interaction.get('headers', MISSING)
        self.body = interaction.get('body', MISSING)
        rules = interaction.get('matchingRules', {})
        if pact.semver['major'] < 2:
            # there are no matchingRules in v1
            self.matching_rules = {}
        elif pact.semver['major'] < 3:
            self.matching_rules = rule_matchers_v2(rules)
        else:
            self.matching_rules = rule_matchers_v3(rules)
        self.result = result

    def log_context(self):
        log.debug(f'Verifying v{self.pact.semver["major"]} Response: status={self.status}, '
                  f'headers={rules_present(self.headers)}, body={rules_present(self.body)}, '
                  f'matching rules={"present" if self.matching_rules != {} else "absent"}')

    def verify(self, response):
        self.log_context()
        if self.status is not MISSING and response.status_code != self.status:
            log.debug(f'.. response {response.text}')
            return self.result.fail(f'{self.interaction_name} status code {response.status_code} is not '
                                    f'expected {self.status}')
        if self.headers is not MISSING:
            for header in self.headers:
                expected = self.headers[header]
                for actual in response.headers:
                    if header.lower() != actual.lower():
                        continue
                    actual = response.headers[actual]
                    # In <v3 the header rules were under $.headers... but in v3 they're {'header': ...} ugh
                    if self.pact.semver['major'] > 2:
                        rule_section = 'header'
                    else:
                        rule_section = 'headers'
                    if not self.check_rules(actual, expected, [rule_section, header]):
                        log.info(f'{self.interaction_name} headers: {response.headers}')
                        return False
                    break
                else:
                    self.result.fail(f'{self.interaction_name} missing header {header!r}')
                    return False

        if self.body is not MISSING:
            if not self.check_rules(response.json(), self.body, ['body']):
                log.info(f'{self.interaction_name} data was: {response.json()}')
                return False
        return True

    def check_rules(self, data, spec, path):
        log.debug(f'check_rules data={data!r} spec={spec!r} path={format_path(path)}')
        if self.matching_rules:
            # if we have matchingRules then just look at those things
            r = self.apply_rules(data, spec, path)
        else:
            # otherwise the actual must equal the expected (excepting dict elements in actual that are unexpected)
            if path[0] in ('header', 'headers'):
                r = self.compare_header(data, spec, path)
            else:
                r = self.compare(data, spec, ['body'])
        log.debug(f'check_rules success={r!r}')
        return r

    def compare_header(self, data, spec, path):
        parsed_data = sorted(parse_header(data))
        parsed_spec = sorted(parse_header(spec))
        log.debug(f'compare_header data={parsed_data} spec={parsed_spec}')
        if parsed_data == parsed_spec:
            return True

        if path[1].lower() != 'content-type':
            return self.result.fail(f'{self.interaction_name} header {path[1]} value {data!r} does not match '
                                    f'expected {spec!r}')
        # there's a specific caveat to header matching that says that if the headers don't match and they're a
        # Content-Type and an encoding present in one and not the other, then that's OK

        # first, confirm the non-charset parts match
        data_without_charset = [part for part in parsed_data if not part.has_param('charset')]
        spec_without_charset = [part for part in parsed_spec if not part.has_param('charset')]
        if data_without_charset != spec_without_charset:
            return self.result.fail(f'{self.interaction_name} header {path[1]} value {data!r} does not match '
                                    f'expected {spec!r} (ignoring charset)')

        # now see whether the presence of the charset differs
        data_has_charset = any(part for part in parsed_data if part.has_param('charset'))
        spec_has_charset = any(part for part in parsed_spec if part.has_param('charset'))
        if data_has_charset == spec_has_charset:
            return self.result.fail(f'{self.interaction_name} header {path[1]} value {data!r} does not match '
                                    f'expected {spec!r}')
        return True

    def compare(self, data, spec, path):
        log.debug(f'compare data={data!r} spec={spec!r} path={format_path(path)}')
        if fold_type(spec) is list:
            return self.compare_list(data, path, spec)
        if fold_type(spec) is dict:
            return self.compare_dict(data, spec, path)
        if not data == spec:
            return self.result.fail(f'Element mismatch {data!r} != {spec!r}', path)
        return True

    def compare_list(self, data, path, spec):
        if fold_type(data) is not list:
            return self.result.fail(f'{self.interaction_name} element is not an array (is {nice_type(data)})', path)
        if len(data) != len(spec):
            return self.result.fail(f'{self.interaction_name} array is incorrect length (is {len(data)} elements)',
                                    path)
        for i, (data_elem, spec_elem) in enumerate(zip(data, spec)):
            p = path + [i]
            if not self.compare(data_elem, spec_elem, p):
                return self.result.fail(f'{self.interaction_name} element {i} ({nice_type(data_elem)}) '
                                        f'does not match spec ({nice_type(spec_elem)})', path)
        return True

    def compare_dict(self, data, spec, path):
        if fold_type(data) is not dict:
            return self.result.fail(f'{self.interaction_name} element is not an object (is {nice_type(data)})', path)
        for key in spec:
            if key not in data:
                return self.result.fail(f'{self.interaction_name} element {key!r} is missing', path)
            p = path + [key]
            if not self.compare(data[key], spec[key], p):
                return self.result.fail(f'{self.interaction_name} element {key} ({nice_type(data[key])}) '
                                        f'does not match spec ({nice_type(spec[key])})', path)
        return True

    def apply_rules(self, data, spec, path):
        # given some actual data and a pact spec at a certain path, check any rules for that path
        log.debug(f'apply_rules data={data!r} spec={spec!r} path={format_path(path)}')
        weighted_rule = self.find_rule(path)
        log.debug(f'... rule lookup got {weighted_rule}')
        if weighted_rule:
            try:
                weighted_rule.rule.apply(data, spec, path)
            except RuleFailed as e:
                log.debug(f'... failed: {e}')
                return self.result.fail(str(e), path)
            log.debug('... passed')

        # we passed the matchingRule but we also need to check the contents of arrays/dicts
        if fold_type(spec) is list:
            r = self.apply_rules_array(data, spec, path)
            log.debug(f'apply_rules {format_path(path)} success={r!r}')
            return r
        elif fold_type(spec) is dict:
            r = self.apply_rules_dict(data, spec, path)
            log.debug(f'apply_rules {format_path(path)} success={r!r}')
            return r
        elif not weighted_rule:
            if path[0] in ('header', 'headers'):
                log.debug('... falling back on header matching')
                return self.compare_header(data, spec, path)
            else:
                # in the absence of a rule, and we're at a leaf, we fall back on equality
                log.debug('... falling back on equality matching')
                return data == spec
        return True

    def find_rule(self, path):
        # rules are trickier to find because the mapping of content to matchingRules isn't 1:1 so...
        section = path[0]
        section_rules = self.matching_rules.get(section)
        if not section_rules:
            return None
        log.debug(f'find_rule got {section_rules} for section {section}')
        if self.pact.semver['major'] > 2:
            # version 3 rules paths don't include the interaction section ("body", "headers", ...)
            path = path[1:]
            if section == 'body':
                # but body paths always have a '$' at the start, yes
                path = ['$'] + path
        else:
            # version 2 paths always have a '$' at the start
            path = ['$'] + path
        weighted_rules = sorted(rule.weight(path) for rule in section_rules)
        display = [(weighted_rule.weight, weighted_rule.rule.path) for weighted_rule in weighted_rules]
        log.debug(f'... path {path} got weights {display}')
        weighted_rule = weighted_rules[-1]
        if weighted_rule.weight:
            return weighted_rule

    def apply_rules_array(self, data, spec, path):
        log.debug(f'apply_rules_array data={data!r} spec={spec!r} path={format_path(path)}')
        if fold_type(data) is not list:
            return self.result.fail(f'{self.interaction_name} element is not an array (is {nice_type(data)})', path)

        if not data and not spec:
            # both arrays are empty, there's no further rules to apply
            return True

        if not spec and data:
            return self.result.fail(f'{self.interaction_name} spec requires empty array but data has contents', path)

        if spec and not data:
            return self.result.fail(f'{self.interaction_name} spec requires data in array but data is empty', path)

        # Attempt to find a matchingRule for the elements in the array
        log.debug('iterating elements, looking for rules')
        for index, data_elem in enumerate(data):
            if not self.apply_rules_array_element(data_elem, spec, path + [index], index):
                log.debug(f'apply_rules_array {path!r} failing on item {index}')
                return False
        log.debug(f'apply_rules_array {path!r} success=True')
        return True

    def apply_rules_array_element(self, data, spec, path, index):
        # we're iterating an array - attempt to find a rule for the specific data element and if
        # the index of the rule is "*" then just use the first element in the spec, otherwise use
        # the *matching* element of the spec.
        log.debug(f'apply_rules_array_element data={data!r} spec={spec!r} path={format_path(path)}')
        weighted_rule = self.find_rule(path)
        log.debug(f'... element rule lookup got {weighted_rule}')
        if len(spec) == 1 and index:
            # if the spec is a single element but data is longer there's a *good chance* that it's a sample
            # for matching rules to be applied to
            spec = spec[0]
        else:
            # element to element comparisons
            spec = spec[index]

        # now do normal rule application
        return self.apply_rules(data, spec, path)

    def apply_rules_dict(self, data, spec, path):
        log.debug(f'apply_rules_dict data={data!r} spec={spec!r} path={format_path(path)}')
        if fold_type(data) is not dict:
            return self.result.fail(f'{self.interaction_name} element is not an object (is {nice_type(data)})', path)
        for k in spec:
            p = path + [k]
            if k not in data:
                # we always flag a failure if a given key is not in the response
                return self.result.fail(f'{self.interaction_name} element {k!r} is missing', path)
            if not self.apply_rules(data[k], spec[k], p):
                return False
        return True


class RequestVerifier(ResponseVerifier):
    interaction_name = 'Request'

    def __init__(self, pact, interaction, result):
        self.method = interaction.get('method', MISSING)
        self.path = interaction.get('path', MISSING)
        self.query = interaction.get('query', MISSING)
        super().__init__(pact, interaction, result)

    def log_context(self):
        log.debug(f'Verifying v{self.pact.semver["major"]} Request: method={self.method}, '
                  f'path={rules_present(self.path)}, query={rules_present(self.query)}, '
                  f'headers={rules_present(self.headers)}, body={rules_present(self.body)}, '
                  f'matching rules={"present" if self.matching_rules != {} else "absent"}')

    def verify(self, request):
        if self.method is not MISSING and request.method.lower() != self.method.lower():
            return self.result.fail(f'Request method {request.method!r} does not match expected {self.method!r}')
        if self.path is not MISSING:
            if self.pact.semver['major'] > 1 and self.matching_rules.get('path'):
                return self.apply_rules(request.path, self.path, ['path'])
            if request.path != self.path:
                return self.result.fail(f'Request path {request.path!r} does not match expected {self.path!r}')
        if self.query is not MISSING:
            if not self.verify_query(self.query, request):
                return False
        return super().verify(request)

    def verify_query(self, spec_query, request):
        if isinstance(spec_query, str):
            spec_query = parse_qs(spec_query)
        request_query = request.query
        if isinstance(request_query, str):
            request_query = parse_qs(request_query)
        if self.pact.semver['major'] > 1 and self.matching_rules.get('query'):
            if not self.apply_rules(request_query, spec_query, ['query']):
                return self.result.fail(f'Request query params {request_query} do not match '
                                        f'expected {spec_query}')
        elif request_query != spec_query:
            return self.result.fail(f'Request query {request_query!r} does not match expected {spec_query!r}')
        return True

    def compare_dict(self, data, spec, path):
        if not super().compare_dict(data, spec, path):
            return False
        # check for unexpected data in the request
        for k in data:
            if k not in spec:
                return self.result.fail(f'Unexpected data in request', path + [k])
        return True


MISSING = 'MISSING'



"""Classes and methods to describe contract Providers."""


class Provider(object):
    """A Pact provider."""

    def __init__(self, name):
        """
        Create a new Provider.

        :param name: The name of this provider. This will be shown in the Pact
            when it is published.
        :type name: str
        """
        self.name = name



import json
import os
import os.path
import urllib.parse

from ..verifier.parse_header import get_header_param
from ..verifier.result import Result
from ..verifier.verify import RequestVerifier


class Request:
    def __init__(self, method, path, query, headers, body):
        self.method = method
        self.path = path
        self.query = query
        self.headers = headers
        self.body = body

    def json(self):
        return self.body


class RecordResult(Result):
    def start(self, interaction):
        super().start(interaction)

    def fail(self, message, path=None):
        self.success = self.FAIL
        self.reason = message
        return not message


class PactVersionConflict(AssertionError):
    pass


class PactInteractionMismatch(AssertionError):
    pass


class PactRequestHandler:
    def __init__(self, pact):
        self.pact = pact

    def validate_request(self, method):
        url_parts = urllib.parse.urlparse(self.path)

        interaction = self.get_interaction(url_parts.path)
        body = self.get_body()

        request = Request(method, url_parts.path, url_parts.query, self.headers, body)
        result = RecordResult()
        RequestVerifier(self.pact, interaction['request'], result).verify(request)
        if not result.success:
            return self.handle_failure(result.reason)
        self.handle_success(interaction)
        if self.pact.file_write_mode != 'never':
            self.write_pact(interaction)
        return self.respond_for_interaction(interaction)

    def get_body(self):
        if not self.body:
            return ''
        content_type = [self.headers[h] for h in self.headers if h.lower() == 'content-type']
        if content_type:
            content_type = content_type[0]
        else:
            # default content type for pacts
            content_type = 'application/json'

        if content_type == 'application/json':
            return json.loads(self.body)
        elif content_type == 'application/x-www-form-urlencoded':
            return urllib.parse.parse_qs(self.body)
        raise ValueError(f'Unhandled body content type {content_type}')

    def get_interaction(self, path):
        raise NotImplementedError()

    def handle_success(self, interaction):
        raise NotImplementedError()

    def handle_failure(self, reason):
        raise NotImplementedError()

    def respond_for_interaction(self, reason):
        raise NotImplementedError()

    def handle_response_encoding(self, response, headers):
        # default to content-type to json
        # rfc4627 states JSON is Unicode and defaults to UTF-8
        content_type = [headers[h] for h in headers if h.lower() == 'content-type']
        if content_type:
            content_type = content_type[0]
            if 'application/json' not in content_type:
                return response['body']
            charset = get_header_param(content_type, 'charset')
            if not charset:
                charset = 'UTF-8'
        else:
            headers['Content-Type'] = 'application/json; charset=UTF-8'
            charset = 'UTF-8'
        return json.dumps(response['body']).encode(charset)

    def write_pact(self, interaction):
        if self.pact.semver["major"] >= 3:
            provider_state_key = 'providerStates'
        else:
            provider_state_key = 'providerState'

        if os.path.exists(self.pact.pact_json_filename):
            with open(self.pact.pact_json_filename) as f:
                pact = json.load(f)
            existing_version = pact['metadata']['pactSpecification']['version']
            if existing_version != self.pact.version:
                raise PactVersionConflict(f'Existing pact ("{pact["interactions"][0]["description"]}") specifies '
                                          f'version {existing_version} but new pact ("interaction["description"]") '
                                          f'specifies version {self.pact.version}')
            for existing in pact['interactions']:
                if (existing['description'] == interaction['description']
                        and existing.get(provider_state_key) == interaction.get(provider_state_key)):
                    # already got one of these...
                    if existing != interaction:
                        raise PactInteractionMismatch(
                            f'Existing "{existing["description"]}" pact given {existing.get(provider_state_key)!r} '
                            'exists with different request/response')
                    return
            pact['interactions'].append(interaction)
        else:
            pact = self.pact.construct_pact(interaction)

        with open(self.pact.pact_json_filename, 'w') as f:
            json.dump(pact, f, indent=2)



from .matchers import (get_generated_values, get_matching_rules_v2,
                       get_matching_rules_v3)


class Request:
    """Represents an HTTP request and supports Matchers on its properties."""

    def __init__(self, method, path, body=None, headers=None, query=''):
        """
        Create a new instance of Request.

        :param method: The HTTP method that is expected.
        :type method: str
        :param path: The URI path that is expected on this request.
        :type path: str, Matcher
        :param body: The contents of the body of the expected request.
        :type body: str, dict, list
        :param headers: The headers of the expected request.
        :type headers: dict
        :param query: The URI query of the expected request.
        :type query: str or dict
        """
        self.method = method
        self.path = path
        self.body = body
        self.headers = headers
        self.query = query

    def json(self, spec_version):
        """Convert the Request to a JSON Pact."""
        request = {
            'method': self.method,
            'path': get_generated_values(self.path)
        }

        if self.headers:
            request['headers'] = get_generated_values(self.headers)

        if self.body is not None:
            request['body'] = get_generated_values(self.body)

        if self.query:
            request['query'] = get_generated_values(self.query)

        if spec_version == '2.0.0':
            matchingRules = self.generate_v2_matchingRules()
        elif spec_version == '3.0.0':
            matchingRules = self.generate_v3_matchingRules()
        else:
            raise ValueError(f'Invalid Pact specification version={spec_version}')

        if matchingRules:
            request['matchingRules'] = matchingRules
        return request

    def generate_v2_matchingRules(self):
        # TODO check there's generation *and* verification tests for all these
        matchingRules = get_matching_rules_v2(self.path, '$.path')
        matchingRules.update(get_matching_rules_v2(self.headers, '$.headers'))
        matchingRules.update(get_matching_rules_v2(self.body, '$.body'))
        matchingRules.update(get_matching_rules_v2(self.query, '$.query'))
        return matchingRules

    def generate_v3_matchingRules(self):
        # TODO check there's generation *and* verification tests for all these
        matchingRules = get_matching_rules_v3(self.path, 'path')
        matchingRules.update(split_header_paths(get_matching_rules_v3(self.headers, 'headers')))

        # body and query rules look different
        body_rules = get_matching_rules_v3(self.body, '$')
        if body_rules:
            matchingRules['body'] = body_rules
        query_rules = get_matching_rules_v3(self.query, 'query')
        if query_rules:
            expand_query_rules(query_rules)
            matchingRules['query'] = query_rules
        return matchingRules


def expand_query_rules(rules):
    # Query rules in the pact JSON are declared without the array notation (even though they always
    # match arrays).
    # The matchers will be coded to JSON paths by get_matching_rules_v3, and we need to extract
    # them out to a dictionary where the original rule path will look like 'query.param'
    # and we need to extract "param".
    # If there's no param (it's just "query") then make it a wildcard
    for rule_path in list(rules):
        matchers = rules.pop(rule_path)
        rule_param = rule_path[6:]
        # trim off any array wildcard, it's implied here
        if rule_param.endswith('[*]'):
            rule_param = rule_param[:-3]
        if not rule_param:
            rule_param = '*'
        rules[rule_param] = matchers


def split_header_paths(rules):
    # Header rules in v3 pacts are stored differently to other types - in a single object called "header"
    # with a sub key per header.
    if not rules:
        return {}
    result = dict(header={})
    for k in rules:
        header = k.split('.')[1]
        result['header'][header] = rules[k]
    return result



import pytest

from pactman.mock.matchers import generate_ruby_protocol, get_generated_values, EachLike, Like, Term, \
    get_matching_rules_v2, get_matching_rules_v3, Equals


def test_generated_value_unknown_type():
    with pytest.raises(ValueError):
        get_generated_values(set())


@pytest.mark.parametrize('input, output', [
    (None, None),
    (False, False),
    ('testing', 'testing'),
    (123, 123),
    (3.14, 3.14),
    ({'id': 123, 'info': {'active': False, 'user': 'admin'}}, {'id': 123, 'info': {'active': False, 'user': 'admin'}}),
    ([1, 123, 'sample'], [1, 123, 'sample']),
    (EachLike({'a': 1}), [{'a': 1}]),
    (EachLike({'a': 1}, minimum=5), [{'a': 1}] * 5),
    (Like(123), 123),
    (Term('[a-f0-9]+', 'abc123'), 'abc123'),
    (Equals(['a', Like('b')]), ['a', 'b']),
    ([EachLike({'username': Term('[a-zA-Z]+', 'firstlast'), 'id': Like(123)})],
     [[{'username': 'firstlast', 'id': 123}]]),
])
def test_generation(input, output):
    assert get_generated_values(input) == output


def test_matching_rules_v2_invald_type():
    with pytest.raises(ValueError):
        assert get_matching_rules_v2(set(), '*')


def test_matching_rules_v3_invald_type():
    with pytest.raises(ValueError):
        assert get_matching_rules_v3(set(), '*')


@pytest.mark.parametrize('input, output', [
    (None, None),
    ('testing', 'testing'),
    (123, 123),
    (3.14, 3.14),
    ({'id': 123, 'info': {'active': False, 'user': 'admin'}}, {'id': 123, 'info': {'active': False, 'user': 'admin'}}),
    ([1, 123, 'sample'], [1, 123, 'sample']),
    (EachLike({'a': 1}), {'json_class': 'Pact::ArrayLike', 'contents': {'a': 1}, 'min': 1}),
    (Like(123), {'json_class': 'Pact::SomethingLike', 'contents': 123}),
    (Term('[a-f0-9]+', 'abc123'), {'json_class': 'Pact::Term',
                                   'data': {'matcher': {'json_class': 'Regexp', 's': '[a-f0-9]+', 'o': 0},
                                            'generate': 'abc123'}}),
    ([EachLike({'username': Term('[a-zA-Z]+', 'firstlast'), 'id': Like(123)})], [
        {'contents': {
            'id': {
                'contents': 123,
                'json_class': 'Pact::SomethingLike'},
            'username': {
                'data': {
                    'generate': 'firstlast',
                    'matcher': {'json_class': 'Regexp', 'o': 0, 's': '[a-zA-Z]+'}},
                'json_class': 'Pact::Term'}},
            'json_class': 'Pact::ArrayLike',
            'min': 1}
    ]),
])
def test_from_term(input, output):
    assert generate_ruby_protocol(input) == output


def test_ruby_protocol_unknown_type():
    with pytest.raises(ValueError):
        generate_ruby_protocol(set())



import os
import tempfile
from unittest import TestCase
from unittest.mock import call, patch

from pactman.mock.consumer import Consumer
from pactman.mock.pact import Pact
from pactman.mock.provider import Provider


class PactTestCase(TestCase):
    def setUp(self):
        self.consumer = Consumer('TestConsumer')
        self.provider = Provider('TestProvider')

    def test_init_defaults(self):
        target = Pact(self.consumer, self.provider)
        self.assertIs(target.consumer, self.consumer)
        self.assertEqual(target.host_name, 'localhost')
        self.assertEqual(target.log_dir, os.getcwd())
        self.assertEqual(target.pact_dir, os.getcwd())
        self.assertEqual(target.port, Pact.BASE_PORT_NUMBER)
        self.assertIs(target.provider, self.provider)
        self.assertIs(target.ssl, False)
        self.assertIsNone(target.sslcert)
        self.assertIsNone(target.sslkey)
        self.assertEqual(target.uri, f'http://localhost:{Pact.BASE_PORT_NUMBER}')
        self.assertEqual(target.version, '2.0.0')
        self.assertEqual(len(target._interactions), 0)

    def test_init_custom_mock_service(self):
        with tempfile.TemporaryDirectory() as d:
            target = Pact(
                self.consumer, self.provider, host_name='192.168.1.1', port=8000,
                log_dir='/logs', ssl=True, sslcert='/ssl.cert', sslkey='/ssl.pem',
                pact_dir=d, version='3.0.0', file_write_mode='merge',
                use_mocking_server=False)

        self.assertIs(target.consumer, self.consumer)
        self.assertEqual(target.host_name, '192.168.1.1')
        self.assertEqual(target.log_dir, '/logs')
        self.assertEqual(target.pact_dir, d)
        self.assertEqual(target.port, 8000)
        self.assertIs(target.provider, self.provider)
        self.assertIs(target.ssl, True)
        self.assertEqual(target.sslcert, '/ssl.cert')
        self.assertEqual(target.sslkey, '/ssl.pem')
        self.assertEqual(target.uri, 'https://192.168.1.1:8000')
        self.assertEqual(target.version, '3.0.0')
        self.assertEqual(target.file_write_mode, 'merge')
        self.assertEqual(len(target._interactions), 0)
        self.assertIs(target.use_mocking_server, False)

    def test_definition_sparse(self):
        target = Pact(self.consumer, self.provider)
        (target
         .given('I am creating a new pact using the Pact class')
         .upon_receiving('a specific request to the server')
         .with_request('GET', '/path')
         .will_respond_with(200, body='success'))

        self.assertEqual(len(target._interactions), 1)

        self.assertEqual(
            target._interactions[0]['providerState'],
            'I am creating a new pact using the Pact class')

        self.assertEqual(
            target._interactions[0]['description'],
            'a specific request to the server')

        self.assertEqual(target._interactions[0]['request'],
                         {'path': '/path', 'method': 'GET'})
        self.assertEqual(target._interactions[0]['response'],
                         {'status': 200, 'body': 'success'})

    def test_definition_all_options(self):
        target = Pact(self.consumer, self.provider, version='2.0.0')
        (target
         .given('I am creating a new pact using the Pact class')
         .upon_receiving('a specific request to the server')
         .with_request('GET', '/path',
                       body={'key': 'value'},
                       headers={'Accept': 'application/json'},
                       query='search=test')
         .will_respond_with(
             200,
             body='success', headers={'Content-Type': 'application/json'}))

        self.assertEqual(
            target._interactions[0]['providerState'],
            'I am creating a new pact using the Pact class')

        self.assertEqual(
            target._interactions[0]['description'],
            'a specific request to the server')

        self.assertEqual(target._interactions[0]['request'], {
            'path': '/path',
            'method': 'GET',
            'body': {'key': 'value'},
            'headers': {'Accept': 'application/json'},
            'query': 'search=test'})
        self.assertEqual(target._interactions[0]['response'], {
            'status': 200,
            'body': 'success',
            'headers': {'Content-Type': 'application/json'}})

    def test_definition_all_options_v3(self):
        target = Pact(self.consumer, self.provider, version='3.0.0')
        (target
         .given([{'name': 'I am creating a new pact using the Pact class', 'params': {}}])
         .upon_receiving('a specific request to the server')
         .with_request('GET', '/path',
                       body={'key': 'value'},
                       headers={'Accept': 'application/json'},
                       query={'search': ['test']})
         .will_respond_with(
             200,
             body='success', headers={'Content-Type': 'application/json'}))

        self.assertEqual(
            target._interactions[0]['providerStates'],
            [{'name': 'I am creating a new pact using the Pact class', 'params': {}}])

        self.assertEqual(
            target._interactions[0]['description'],
            'a specific request to the server')

        self.assertEqual(target._interactions[0]['request'], {
            'path': '/path',
            'method': 'GET',
            'body': {'key': 'value'},
            'headers': {'Accept': 'application/json'},
            'query': {'search': ['test']}})
        self.assertEqual(target._interactions[0]['response'], {
            'status': 200,
            'body': 'success',
            'headers': {'Content-Type': 'application/json'}})

    def test_definition_v3_requires_new_providerStates(self):
        target = Pact(self.consumer, self.provider, version='3.0.0')
        target.given('I am creating a new pact using the Pact class')
        self.assertEqual(
            target._interactions[0]['providerStates'],
            [{'name': 'I am creating a new pact using the Pact class', 'params': {}}])

    def test_definition_multiple_interactions(self):
        target = Pact(self.consumer, self.provider)
        (target
         .given('I am creating a new pact using the Pact class')
         .upon_receiving('a specific request to the server')
         .with_request('GET', '/foo')
         .will_respond_with(200, body='success')
         .given('I am creating another new pact using the Pact class')
         .upon_receiving('a different request to the server')
         .with_request('GET', '/bar')
         .will_respond_with(200, body='success'))

        self.assertEqual(len(target._interactions), 2)

        self.assertEqual(
            target._interactions[1]['providerState'],
            'I am creating a new pact using the Pact class')
        self.assertEqual(
            target._interactions[0]['providerState'],
            'I am creating another new pact using the Pact class')

        self.assertEqual(
            target._interactions[1]['description'],
            'a specific request to the server')
        self.assertEqual(
            target._interactions[0]['description'],
            'a different request to the server')

        self.assertEqual(target._interactions[1]['request'],
                         {'path': '/foo', 'method': 'GET'})
        self.assertEqual(target._interactions[0]['request'],
                         {'path': '/bar', 'method': 'GET'})

        self.assertEqual(target._interactions[1]['response'],
                         {'status': 200, 'body': 'success'})
        self.assertEqual(target._interactions[0]['response'],
                         {'status': 200, 'body': 'success'})


class PactSetupTestCase(PactTestCase):
    def setUp(self):
        super(PactSetupTestCase, self).setUp()
        self.addCleanup(patch.stopall)
        self.target = Pact(self.consumer, self.provider)
        (self.target
         .given('I am creating a new pact using the Pact class')
         .upon_receiving('a specific request to the server')
         .with_request('GET', '/path')
         .will_respond_with(200, body='success'))

        self.delete_call = call('delete', 'http://localhost:1234/interactions',
                                headers={'X-Pact-Mock-Service': 'true'})

        self.put_interactions_call = call(
            'put', 'http://localhost:1234/interactions',
            data=None,
            headers={'X-Pact-Mock-Service': 'true'},
            json={'interactions': [{
                'response': {'status': 200, 'body': 'success'},
                'request': {'path': '/path', 'method': 'GET'},
                'description': 'a specific request to the server',
                'providerState': 'I am creating a new pact using the Pact class'}]})


class PactContextManagerTestCase(PactTestCase):
    def setUp(self):
        super(PactContextManagerTestCase, self).setUp()
        self.addCleanup(patch.stopall)
        self.mock_setup = patch.object(
            Pact, 'setup', autospec=True).start()

        self.mock_verify = patch.object(
            Pact, 'verify', autospec=True).start()

    def test_successful(self):
        pact = Pact(self.consumer, self.provider)
        with pact:
            pass

        self.mock_setup.assert_called_once_with(pact)
        self.mock_verify.assert_called_once_with(pact)

    def test_context_raises_error(self):
        pact = Pact(self.consumer, self.provider)
        with self.assertRaises(RuntimeError):
            with pact:
                raise RuntimeError

        self.mock_setup.assert_called_once_with(pact)
        self.assertFalse(self.mock_verify.called)



import collections
from unittest.mock import Mock

import pytest
from pactman.verifier.matching_rule import (InvalidMatcher, MatchDecimal,
                                            MatchEquality, Matcher,
                                            MatchInclude, MatchInteger,
                                            MatchNull, MatchNumber, MatchRegex,
                                            MatchType, RuleFailed, fold_type,
                                            log, rule_matchers_v2,
                                            rule_matchers_v3, split_path,
                                            weight_path)


def test_stringify():
    r = MatchType('$', {'match': 'type'})
    assert str(r) == "Rule match by {'match': 'type'} at $"
    assert repr(r) == "<MatchType path='$' rule={'match': 'type'}>"


def test_invalid_match_type(monkeypatch):
    monkeypatch.setattr(log, 'warning', Mock())
    assert isinstance(Matcher.get_matcher('$', {'match': 'spam'}), InvalidMatcher)
    log.warning.assert_called_once()


@pytest.mark.parametrize('path, weight, spam_weight', [
    ('$.body', 4, 4),
    ('$.body.item1.level[0].id', 0, 0),
    ('$.body.item1.level[1].id', 64, 0),
    ('$.body.item1.level[*].id', 32, 0),
    ('$.body.*.level[*].id', 16, 0),
    ('$.body.*.*[*].id', 8, 8),
])
def test_weightings(path, weight, spam_weight):
    rule = Matcher(path, {'match': 'type'})
    assert rule.weight(['$', 'body', 'item1', 'level', 1, 'id']).weight == weight
    assert rule.weight(['$', 'body', 'item2', 'spam', 1, 'id']).weight == spam_weight


@pytest.mark.parametrize('path, result', [
    ('', []),
    ('$', ['$']),
    ('$.body', ['$', 'body']),
    ('$.body.item1', ['$', 'body', 'item1']),
    ('$.body.item2', ['$', 'body', 'item2']),
    ('$.header.item1', ['$', 'header', 'item1']),
    ('$.body.item1.level', ['$', 'body', 'item1', 'level']),
    ('$.body.item1.level[1]', ['$', 'body', 'item1', 'level', 1]),
    ('$.body.item1.level[1].id', ['$', 'body', 'item1', 'level', 1, 'id']),
    ('$.body.item1.level[1].name', ['$', 'body', 'item1', 'level', 1, 'name']),
    ('$.body.item1.level[2]', ['$', 'body', 'item1', 'level', 2]),
    ('$.body.item1.level[2].id', ['$', 'body', 'item1', 'level', 2, 'id']),
    ('$.body.item1.level[*].id', ['$', 'body', 'item1', 'level', '*', 'id']),
    ('$.body.*.level[*].id', ['$', 'body', '*', 'level', '*', 'id']),
])
def test_split_path(path, result):
    assert list(split_path(path)) == result


@pytest.mark.parametrize('spec, test, result', [
    (['a'], ['b'], 0),
    (['*'], ['a'], 1),
    (['a', '*'], ['a', 'b'], 2),
    (['a', 'b'], ['a', 'b'], 4),
    (['a', 'b', 'c'], ['a', 'b'], 0),
    (['a', 'b'], ['a', 'b', 'c'], 4),
])
def test_weight_path(spec, test, result):
    assert weight_path(spec, test) == result


@pytest.mark.parametrize('data, spec', [
    (1, 1),
    (1, 1.0),
    (1.0, 1.0),
    (1.0, 1.0),
])
def test_numbers(data, spec):
    MatchType('$', dict(match='type')).apply(data, spec, ['a'])


def test_regex():
    MatchRegex('$', dict(match='regex', regex=r'\w+')).apply('spam', None, ['a'])


def test_regex_fail():
    with pytest.raises(RuleFailed):
        MatchRegex('$', dict(match='regex', regex=r'\W+')).apply('spam', None, ['a'])


def test_integer():
    MatchInteger('$', dict(match='integer')).apply(1, None, ['a'])


def test_integer_fail():
    with pytest.raises(RuleFailed):
        MatchInteger('$', dict(match='integer')).apply(1.0, None, ['a'])


def test_decimal():
    MatchDecimal('$', dict(match='decimal')).apply(1.0, None, ['a'])


def test_decimal_fail():
    with pytest.raises(RuleFailed):
        MatchDecimal('$', dict(match='decimal')).apply(1, None, ['a'])


@pytest.mark.parametrize('value', [1, 1.0])
def test_number(value):
    MatchNumber('$', dict(match='number')).apply(value, None, ['a'])


def test_number_fail():
    with pytest.raises(RuleFailed):
        MatchNumber('$', dict(match='number')).apply('spam', None, ['a'])


def test_equality():
    MatchEquality('$', dict(match='equality')).apply('spam', 'spam', ['a'])


def test_equality_fail():
    with pytest.raises(RuleFailed):
        MatchEquality('$', dict(match='equality')).apply('ham', 'spam', ['a'])


def test_include():
    MatchInclude('$', dict(match='include', value='spam')).apply('spammer', None, ['a'])


def test_include_fail():
    with pytest.raises(RuleFailed):
        MatchInclude('$', dict(match='include', value='spam')).apply('ham', None, ['a'])


def test_null():
    MatchNull('$', dict(match='null')).apply(None, None, ['a'])


def test_null_fail():
    with pytest.raises(RuleFailed):
        MatchNull('$', dict(match='null')).apply('ham', None, ['spam'])


def test_min():
    MatchType('$', dict(match='type', min=1)).apply(['spam'], ['a'], [])


def test_min_not_met():
    with pytest.raises(RuleFailed):
        MatchType('$', dict(match='type', min=2)).apply(['spam'], ['a', 'b'], ['spam'])


def test_min_ignored():
    MatchType('$', dict(match='type', min=1)).apply(0, 0, [])


def test_max():
    MatchType('$', dict(match='type', max=1)).apply(['spam'], ['a'], [])


def test_max_not_met():
    with pytest.raises(RuleFailed):
        MatchType('$', dict(match='type', max=2)).apply([1, 2, 3], [1, 2], ['spam'])


def test_max_ignored():
    MatchType('$', dict(match='type', max=1)).apply(0, 0, [])


@pytest.mark.parametrize('source, result', [
    ({}, dict),
    ([], list),
    (collections.OrderedDict(), dict),
])
def test_fold_type(source, result):
    assert fold_type(source) == result


def test_build_matching_rules_handles_rule_with_unknown_type_v2(monkeypatch):
    monkeypatch.setattr(log, 'warning', Mock())
    rules = rule_matchers_v2({
        "$.body": {"match": "SPAM"},
        "$.body[*].*": {"match": "type"}
    })
    assert 2 == len(rules['body'])
    log.warning.assert_called_once()


def test_build_matching_rules_handles_rule_with_unknown_type_v3(monkeypatch):
    monkeypatch.setattr(log, 'warning', Mock())
    rules = rule_matchers_v3({
        "body": {
            "$": {"matchers": [{"match": "SPAM"}]},
            "$[*].*": {"matchers": [{"match": "type"}]}
        }
    })
    assert 2 == len(rules['body'])
    log.warning.assert_called_once()



import pytest

from pactman import SomethingLike, Like
from pactman.mock.matchers import Matcher, Term


def test_is_something_like():
    assert SomethingLike is Like


def test_valid_types():
    types = [None, list(), dict(), 1, 1.0, 'string', 'unicode', Matcher()]

    for t in types:
        SomethingLike(t)


def test_invalid_types():
    with pytest.raises(AssertionError) as e:
        SomethingLike(set())

    assert 'matcher must be one of ' in str(e.value)


def test_basic_type():
    assert SomethingLike(123).ruby_protocol() == {'json_class': 'Pact::SomethingLike', 'contents': 123}


def test_complex_type():
    assert SomethingLike({'name': Term('.+', 'admin')}).ruby_protocol() == {
        'json_class': 'Pact::SomethingLike',
        'contents': {'name': {
            'json_class': 'Pact::Term',
            'data': {
                'matcher': {
                    'json_class': 'Regexp',
                    's': '.+',
                    'o': 0
                },
                'generate': 'admin'
            }
        }}
    }



from pactman import Term
from pactman.mock.request import Request


def test_matcher_in_path_gets_converted():
    target = Request('GET', '/', headers={'Spam': Term(r'\w+', 'spam')})
    assert target.json('3.0.0') == {
        'method': 'GET',
        'path': '/',
        'headers': {'Spam': 'spam'},
        'matchingRules': {
            'header': {
                "Spam": {
                    "matchers": [
                        {
                            "match": "regex",
                            "regex": "\\w+"
                        }
                    ]
                }
            }
        }
    }



import pytest

from pactman import Consumer, Like, Provider, Term
from pactman.mock.request import Request


def test_v2():
    pact = Consumer('consumer').has_pact_with(Provider('provider'), version='2.0.0')

    pact.given("the condition exists").upon_receiving("a request") \
        .with_request("GET", "/path", query="fields=first,second").will_respond_with(200, body='ok')

    result = pact.construct_pact(pact._interactions[0])
    assert result == {
        'consumer': {'name': 'consumer'},
        'provider': {'name': 'provider'},
        'interactions': [
            {
                'description': 'a request',
                'providerState': 'the condition exists',
                'request': dict(method='GET', path='/path', query='fields=first,second'),
                'response': dict(status=200, body='ok'),

            }
        ],
        'metadata': dict(pactSpecification=dict(version='2.0.0'))
    }


@pytest.mark.parametrize('query_field', [
    'first,second',
    ['first,second']
])
def test_v3(query_field):
    pact = Consumer('consumer').has_pact_with(Provider('provider'), version='3.0.0')
    pact.given([{'name': "the condition exists", 'params': {}}]).upon_receiving("a request") \
        .with_request("GET", "/path", query=dict(fields=query_field)).will_respond_with(200, body='ok')

    result = pact.construct_pact(pact._interactions[0])
    assert result == {
        'consumer': {'name': 'consumer'},
        'provider': {'name': 'provider'},
        'interactions': [
            {
                'description': 'a request',
                'providerStates': [{'name': 'the condition exists', 'params': {}}],
                'request': dict(method='GET', path='/path', query=dict(fields=['first,second'])),
                'response': dict(status=200, body='ok'),

            }
        ],
        'metadata': dict(pactSpecification=dict(version='3.0.0'))
    }


def test_like_v2():
    pact = Consumer('consumer').has_pact_with(Provider('provider'), version='2.0.0')

    pact.given("the condition exists").upon_receiving("a request") \
        .with_request("GET", "/path", query=Like("fields=first,second")).will_respond_with(200, body='ok')

    result = pact.construct_pact(pact._interactions[0])
    assert result == {
        'consumer': {'name': 'consumer'},
        'provider': {'name': 'provider'},
        'interactions': [
            {
                'description': 'a request',
                'providerState': 'the condition exists',
                'request': dict(method='GET', path='/path', query='fields=first,second',
                                matchingRules={'$.query': {'match': 'type'}}),
                'response': dict(status=200, body='ok'),

            }
        ],
        'metadata': dict(pactSpecification=dict(version='2.0.0'))
    }


def test_like_v3():
    pact = (
        Consumer('consumer').has_pact_with(Provider('provider'), version='3.0.0')
        .given("the condition exists")
        .upon_receiving("a request")
        .with_request("GET", "/path", query=dict(fields=Like(['first,second'])))
        .will_respond_with(200, body='ok')
    )

    result = pact.construct_pact(pact._interactions[0])
    assert result == {
        'consumer': {'name': 'consumer'},
        'provider': {'name': 'provider'},
        'interactions': [
            {
                'description': 'a request',
                'providerStates': [{'name': 'the condition exists', 'params': {}}],
                'request': dict(method='GET', path='/path', query=dict(fields=['first,second']),
                                matchingRules={'query': {'fields': {'matchers': [{'match': 'type'}]}}}),
                'response': dict(status=200, body='ok'),

            }
        ],
        'metadata': dict(pactSpecification=dict(version='3.0.0'))
    }


def test_broader_like_v3():
    pact = (
        Consumer('consumer').has_pact_with(Provider('provider'), version='3.0.0')
        .given("the condition exists")
        .upon_receiving("a request")
        .with_request("GET", "/path", query=Like(dict(fields=['first,second'])))
        .will_respond_with(200, body='ok')
    )

    result = pact.construct_pact(pact._interactions[0])
    assert result == {
        'consumer': {'name': 'consumer'},
        'provider': {'name': 'provider'},
        'interactions': [
            {
                'description': 'a request',
                'providerStates': [{'name': 'the condition exists', 'params': {}}],
                'request': dict(method='GET', path='/path', query=dict(fields=['first,second']),
                                matchingRules={'query': {'*': {'matchers': [{'match': 'type'}]}}}),
                'response': dict(status=200, body='ok'),
            }
        ],
        'metadata': dict(pactSpecification=dict(version='3.0.0'))
    }


def test_matcher_in_query():
    target = Request('GET', '/test-path', query={'q': [Like('spam')], 'l': [Term(r'\d+', '10')]})
    assert target.json('3.0.0') == {
        'method': 'GET',
        'path': '/test-path',
        'query': {'q': ['spam'], 'l': ['10']},
        'matchingRules': {
            'query': {
                'q': {
                    'matchers': [
                        {
                            'match': 'type'
                        },
                    ]
                },
                'l': {
                    'matchers': [
                        {
                            'match': 'regex',
                            'regex': r'\d+',
                        }
                    ]
                },
            }
        }
    }



from pactman import Term
from pactman.mock.request import Request


def test_matcher_in_path_gets_converted():
    target = Request('GET', Term(r'\/.+', '/test-path'))
    assert target.json('2.0.0') == {
        'method': 'GET',
        'path': '/test-path',
        'matchingRules': {
            '$.path': {
                'regex': r'\/.+'
            }
        }
    }



"""API for creating a contract and configuring the mock service."""
from __future__ import unicode_literals

import os

import semver

from ..mock.request import Request
from ..mock.response import Response
from .mock_server import getMockServer
from .mock_urlopen import MockURLOpenHandler

USE_MOCKING_SERVER = os.environ.get('PACT_USE_MOCKING_SERVER', 'no') == 'yes'


def ensure_pact_dir(pact_dir):
    if not os.path.exists(pact_dir):
        parent_dir = os.path.dirname(pact_dir)
        if not os.path.exists(parent_dir):
            raise ValueError(f'Pact destination directory {pact_dir} does not exist')
        os.mkdir(pact_dir)


class Pact(object):
    """
    Represents a contract between a consumer and provider.

    Provides Python context handlers to configure the Pact mock service to
    perform tests on a Python consumer. For example:

    >>> from pactman import Consumer, Provider
    >>> pact = Consumer('consumer').has_pact_with(Provider('provider'))
    >>> (pact.given('the echo service is available')
    ...  .upon_receiving('a request is made to the echo service')
    ...  .with_request('get', '/echo', query={'text': 'Hello!'})
    ...  .will_respond_with(200, body='Hello!'))
    >>> with pact:
    ...   requests.get(pact.uri + '/echo?text=Hello!')

    The GET request is made to the mock service, which will verify that it
    was a GET to /echo with a query string with a key named `text` and its
    value is `Hello!`. If the request does not match an error is raised, if it
    does match the defined interaction, it will respond with the text `Hello!`.
    """

    HEADERS = {'X-Pact-Mock-Service': 'true'}

    def __init__(self, consumer, provider, host_name='localhost', port=None,
                 log_dir=None, ssl=False, sslcert=None, sslkey=None,
                 pact_dir=None, version='2.0.0',
                 file_write_mode='overwrite', use_mocking_server=USE_MOCKING_SERVER):
        """
        Constructor for Pact.

        :param consumer: The consumer for this contract.
        :type consumer: pact.Consumer
        :param provider: The provider for this contract.
        :type provider: pact.Provider
        :param host_name: The host name where the mock service is running.
        :type host_name: str
        :param port: The port number where the mock service is running. Defaults
            to a port >= 8050.
        :type port: int
        :param log_dir: The directory where logs should be written. Defaults to
            the current directory.
        :type log_dir: str
        :param ssl: Flag to control the use of a self-signed SSL cert to run
            the server over HTTPS , defaults to False.
        :type ssl: bool
        :param sslcert: Path to a custom self-signed SSL cert file, 'ssl'
            option must be set to True to use this option. Defaults to None.
        :type sslcert: str
        :param sslkey: Path to a custom key and self-signed SSL cert key file,
            'ssl' option must be set to True to use this option.
            Defaults to None.
        :type sslkey: str
        :param pact_dir: Directory where the resulting pact files will be
            written. Defaults to the current directory.
        :type pact_dir: str
        :param version: The Pact Specification version to use, defaults to
            '2.0.0'.
        :type version: str
        :param file_write_mode: `overwrite` or `merge`. Use `merge` when
            running multiple mock service instances in parallel for the same
            consumer/provider pair. Ensure the pact file is deleted before
            running tests when using this option so that interactions deleted
            from the code are not maintained in the file. Defaults to
            `overwrite`.
        :type file_write_mode: str
        :param use_mocking_server: If True the mocking will be done using a
            HTTP server rather than patching urllib3 connections.
        :type use_mocking_server: bool
        """
        self.scheme = 'https' if ssl else 'http'
        self.consumer = consumer
        self.file_write_mode = file_write_mode
        self.host_name = host_name
        self.log_dir = log_dir or os.getcwd()
        self.pact_dir = pact_dir or os.getcwd()
        self.port = port or self.allocate_port()
        self.provider = provider
        # TODO implement SSL
        self.ssl = ssl
        self.sslcert = sslcert
        self.sslkey = sslkey
        self.version = version
        self.semver = semver.parse(self.version)
        self.use_mocking_server = use_mocking_server
        self._interactions = []
        self._mock_handler = None

        self.check_existing_file()

    @property
    def uri(self):
        return '{scheme}://{host_name}:{port}'.format(host_name=self.host_name, port=self.port, scheme=self.scheme)

    BASE_PORT_NUMBER = 8150

    @classmethod
    def allocate_port(cls):
        cls.BASE_PORT_NUMBER += 5
        return cls.BASE_PORT_NUMBER

    def check_existing_file(self):
        # ensure destination directory exists
        if self.file_write_mode == 'never':
            return
        ensure_pact_dir(self.pact_dir)
        if self.file_write_mode == 'overwrite':
            if os.path.exists(self.pact_json_filename):
                os.remove(self.pact_json_filename)

    @property
    def pact_json_filename(self):
        return os.path.join(self.pact_dir, f'{self.consumer.name}-{self.provider.name}-pact.json')

    def given(self, provider_state, **params):
        """
        Define the provider state for this pact.

        When the provider verifies this contract, they will use this field to
        setup pre-defined data that will satisfy the response expectations.

        In pact v2 the provider state is a short sentence that is unique to describe
        the provider state for this contract. For example:

            "an alligator with the given name Mary exists and the spam nozzle is operating"

        In pact v3 the provider state is a list of state specifications with a name and
        associated params to define specific values for the state. This may be provided
        in two ways. Either call with a single list, for example:

            [
                {
                    "name": "an alligator with the given name exists",
                    "params": {"name" : "Mary"}
                }, {
                    "name": "the spam nozzle is operating",
                    "params" : {}
                }
            ]

        or for convenience call `.given()` with a string as in v2, which implies a single
        provider state, with params taken from keyword arguments like so:

            .given("an alligator with the given name exists", name="Mary")

        If additional provider states are required for a v3 pact you may either use the list
        form above, or make an additional call to `.and_given()`.

        If you don't have control over the provider, and they cannot implement a provider
        state, you may use an explicit `None` for the provider state value. This is
        discouraged as it introduces fragile external dependencies in your tests.

        :param provider_state: The state as described above.
        :type provider_state: string or list as above
        :rtype: Pact
        """
        if provider_state is None:
            self._interactions.insert(0, {})
            return self

        if self.semver["major"] < 3:
            provider_state_key = 'providerState'
            if not isinstance(provider_state, str):
                raise ValueError('pact v2 provider states must be strings')
        else:
            provider_state_key = 'providerStates'
            if isinstance(provider_state, str):
                provider_state = [{'name': provider_state, 'params': params}]
            elif not isinstance(provider_state, list):
                raise ValueError('pact v3+ provider states must be lists of {name: "", params: {}} specs')
        self._interactions.insert(0, {provider_state_key: provider_state})
        return self

    def and_given(self, provider_state, **params):
        """
        Define an additional provider state for this pact.

        Supply the provider state name and any params taken in keyword arguments like so:

            .given("an alligator with the given name exists", name="Mary")

        :param provider_state: The state as described above.
        :type provider_state: string or list as above
        :rtype: Pact
        """
        if self.semver["major"] < 3:
            raise ValueError('pact v2 only allows a single provider state')
        elif not self._interactions:
            raise ValueError('only invoke and_given() after given()')
        self._interactions[-1]['providerStates'].append({'name': provider_state, 'params': params})
        return self

    def setup(self):
        self._mock_handler.setup(self._interactions)

    def start_mocking(self):
        if self.use_mocking_server:
            self._mock_handler = getMockServer(self)
        else:
            # ain't no port, we're monkey-patching (but the URLs we generate still need to look correct)
            self._mock_handler = MockURLOpenHandler(self)

    def stop_mocking(self):
        self._mock_handler.terminate()
        self._mock_handler = None

    # legacy pact-python API support
    start_service = start_mocking
    stop_service = stop_mocking

    def upon_receiving(self, scenario):
        """
        Define the name of this contract.

        :param scenario: A unique name for this contract.
        :type scenario: basestring
        :rtype: Pact
        """
        self._interactions[0]['description'] = scenario
        return self

    def verify(self):
        """
        Have the mock service verify all interactions occurred.

        Calls the mock service to verify that all interactions occurred as
        expected, and has it write out the contracts to disk.

        :raises AssertionError: When not all interactions are found.
        """
        try:
            self._mock_handler.verify()
        finally:
            # clear the interactions once we've attempted to verify, allowing re-use of the mock
            self._interactions[:] = []

    def with_request(self, method, path, body=None, headers=None, query=None):
        """
        Define the request the request that the client is expected to perform.

        :param method: The HTTP method.
        :type method: str
        :param path: The path portion of the URI the client will access.
        :type path: str, Matcher
        :param body: The request body, can be a string or an object that will
            serialize to JSON, like list or dict, defaults to None.
        :type body: list, dict or None
        :param headers: The headers the client is expected to include on with
            this request. Defaults to None.
        :type headers: dict or None
        :param query: The query options the client is expected to send. Can be
            a dict of keys and values, or a URL encoded string.
            Defaults to None.
        :type query: dict, str, or None
        :rtype: Pact
        """
        # ensure all query values are lists of values
        if isinstance(query, dict):
            for k, v in query.items():
                if isinstance(v, str):
                    query[k] = [v]
        self._interactions[0]['request'] = Request(
            method, path, body=body, headers=headers, query=query).json(self.version)
        return self

    def will_respond_with(self, status, headers=None, body=None):
        """
        Define the response the server is expected to create.

        :param status: The HTTP status code.
        :type status: int
        :param headers: All required headers. Defaults to None.
        :type headers: dict or None
        :param body: The response body, or a collection of Matcher objects to
            allow for pattern matching. Defaults to None.
        :type body: Matcher, dict, list, basestring, or None
        :rtype: Pact
        """
        self._interactions[0]['response'] = Response(status,
                                                     headers=headers,
                                                     body=body).json(self.version)
        return self

    _auto_mocked = False

    def __enter__(self):
        """
        Handler for entering a Python context.

        Sets up the mock service to expect the client requests.
        """
        if not self.use_mocking_server and not self._mock_handler:
            self._auto_mocked = True
            self.start_mocking()

        self.setup()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Handler for exiting a Python context.

        Calls the mock service to verify that all interactions occurred as
        expected, and has it write out the contracts to disk.
        """
        if (exc_type, exc_val, exc_tb) != (None, None, None):
            return

        self.verify()

        if not self.use_mocking_server and self._auto_mocked:
            self.stop_mocking()

    def construct_pact(self, interaction):
        """Construct a pact JSON data structure for the interaction.
        """
        return dict(
            consumer={"name": self.consumer.name},
            provider={"name": self.provider.name},
            interactions=[interaction],
            metadata=dict(pactSpecification=dict(version=self.version)),
        )



import requests

from pactman import Pact


def test_params_url_coded():
    pact = Pact('Consumer', 'Provider', file_write_mode='never').given('everything is ideal') \
        .upon_receiving('a request').with_request(
            method='POST',
            path="/",
            headers={'Content-Type': 'application/x-www-form-urlencoded'},
            query={
                 "client_id": ["test1"],
                 "secret": ["test1-secret-key"],
                 "scope": ['openid', 'profile', 'phone', 'offline_access']
            }
        ).will_respond_with(200, body='some data')
    with pact:
        requests.post(pact.uri, params=dict(client_id='test1', secret='test1-secret-key',
                                            scope='openid profile phone offline_access'.split()),
                      headers={'Content-Type': 'application/x-www-form-urlencoded'})


def test_body_url_coded():
    pact = Pact('Consumer', 'Provider', file_write_mode='never').given('everything is ideal') \
        .upon_receiving('a request').with_request(
             method='POST',
             path="/",
             headers={'Content-Type': 'application/x-www-form-urlencoded'},
             body={
                 "client_id": ["test1"],
                 "secret": ["test1-secret-key"],
                 "scope": ['openid', 'profile', 'phone', 'offline_access']
             }
        ).will_respond_with(200, body='some data')
    with pact:
        requests.post(pact.uri, data=dict(client_id='test1', secret='test1-secret-key',
                                          scope='openid profile phone offline_access'.split()),
                      headers={'Content-Type': 'application/x-www-form-urlencoded'})



import glob
import logging
import pytest
import os
from _pytest.outcomes import Failed

from .broker_pact import BrokerPact, BrokerPacts
from .result import log, PytestResult


def pytest_addoption(parser):
    parser.addoption("--pact-files", default=None,
                     help="pact JSON files to verify (wildcards allowed)")
    parser.addoption("--pact-broker-url", default='',
                     help="pact broker URL")
    parser.addoption("--pact-provider-name", default=None,
                     help="provider pact name")
    parser.addoption("--pact-publish-results", action="store_true", default=False,
                     help="report pact results to pact broker")
    parser.addoption("--pact-provider-version", default=None,
                     help="provider version to use when reporting pact results to pact broker")


def get_broker_url(config):
    return config.getoption('pact_broker_url') or os.environ.get('PACT_BROKER_URL')


# add the pact broker URL to the pytest output if running verbose
def pytest_report_header(config):
    if config.getoption('verbose') > 0:
        location = get_broker_url(config) or config.getoption('pact_files')
        return [f'Loading pacts from {location}']


def pytest_configure(config):
    logging.getLogger('pactman').handlers = []
    logging.basicConfig(format='%(message)s')
    verbosity = config.getoption('verbose')
    if verbosity > 0:
        log.setLevel(logging.DEBUG)


class PytestPactVerifier:
    def __init__(self, publish_results, provider_version, interaction, consumer):
        self.publish_results = publish_results
        self.provider_version = provider_version
        self.interaction = interaction
        self.consumer = consumer

    def verify(self, provider_url, provider_setup):
        try:
            self.interaction.verify_with_callable_setup(provider_url, provider_setup)
        except (Failed, AssertionError) as e:
            raise Failed(str(e)) from None

    def finish(self):
        if self.consumer and self.publish_results and self.provider_version:
            self.consumer.publish_result(self.provider_version)


def flatten_pacts(pacts):
    for consumer in pacts:
        last = consumer.interactions[-1]
        for interaction in consumer.interactions:
            if interaction is last:
                yield (interaction, consumer)
            else:
                yield (interaction, None)


def get_pact_files(file_location):
    if not file_location:
        return []
    for filename in glob.glob(file_location):
        yield BrokerPact.load_file(filename, result_factory=PytestResult)


def test_id(identifier):
    interaction, _ = identifier
    return str(interaction)


def pytest_generate_tests(metafunc):
    if 'pact_verifier' in metafunc.fixturenames:
        broker_url = get_broker_url(metafunc.config)
        if not broker_url:
            pact_files = get_pact_files(metafunc.config.getoption('pact_files'))
            if not pact_files:
                raise ValueError('need a --pact-broker-url or --pact-files option')
            metafunc.parametrize("pact_verifier", flatten_pacts(pact_files), ids=test_id, indirect=True)
        else:
            provider_name = metafunc.config.getoption('pact_provider_name')
            if not provider_name:
                raise ValueError('--pact-broker-url requires the --pact-provider-name option')
            broker_pacts = BrokerPacts(provider_name, pact_broker_url=broker_url, result_factory=PytestResult)
            metafunc.parametrize("pact_verifier", flatten_pacts(broker_pacts.consumers()),
                                 ids=test_id, indirect=True)


@pytest.fixture()
def pact_verifier(pytestconfig, request):
    interaction, consumer = request.param
    p = PytestPactVerifier(pytestconfig.getoption('pact_publish_results'),
                           pytestconfig.getoption('pact_provider_version'),
                           interaction, consumer)
    yield p
    p.finish()



from pactman import Term


def test_regex():
    assert Term('[a-zA-Z]', 'abcXYZ').ruby_protocol() == {
        'json_class': 'Pact::Term',
        'data': {
            'matcher': {
                'json_class': 'Regexp',
                's': '[a-zA-Z]',
                'o': 0},
            'generate': 'abcXYZ'}
    }



import pytest

from pactman import Consumer, Provider


@pytest.mark.parametrize('version', ['2.0.0', '3.0.0'])
def test_null(version):
    pact = Consumer('consumer').has_pact_with(Provider('provider'), version=version)
    pact.given(None).upon_receiving("a request").with_request("GET", "/path") \
        .will_respond_with(200, body='ok')

    result = pact.construct_pact(pact._interactions[0])
    assert result == {
        'consumer': {'name': 'consumer'},
        'provider': {'name': 'provider'},
        'interactions': [
            {
                'description': 'a request',
                'request': dict(method='GET', path='/path'),
                'response': dict(status=200, body='ok'),

            }
        ],
        'metadata': dict(pactSpecification=dict(version=version))
    }


def test_v2():
    pact = Consumer('consumer').has_pact_with(Provider('provider'), version='2.0.0')
    pact.given("the condition exists").upon_receiving("a request").with_request("GET", "/path") \
        .will_respond_with(200, body='ok')

    result = pact.construct_pact(pact._interactions[0])
    assert result == {
        'consumer': {'name': 'consumer'},
        'provider': {'name': 'provider'},
        'interactions': [
            {
                'description': 'a request',
                'providerState': 'the condition exists',
                'request': dict(method='GET', path='/path'),
                'response': dict(status=200, body='ok'),

            }
        ],
        'metadata': dict(pactSpecification=dict(version='2.0.0'))
    }


def test_v3():
    pact = Consumer('consumer').has_pact_with(Provider('provider'), version='3.0.0')
    pact.given([
        dict(name="the condition exists", params={}),
        dict(name="the user exists", params=dict(username='alex')),
    ]).upon_receiving("a request").with_request("GET", "/path").will_respond_with(200, body='ok')

    result = pact.construct_pact(pact._interactions[0])
    assert result == {
        'consumer': {'name': 'consumer'},
        'provider': {'name': 'provider'},
        'interactions': [
            {
                'description': 'a request',
                'providerStates': [
                    {'name': 'the condition exists', 'params': {}},
                    {'name': 'the user exists', 'params': {'username': 'alex'}}
                ],
                'request': dict(method='GET', path='/path'),
                'response': dict(status=200, body='ok'),

            }
        ],
        'metadata': dict(pactSpecification=dict(version='3.0.0'))
    }


def test_v3_and_given():
    pact = (
        Consumer('consumer').has_pact_with(Provider('provider'), version='3.0.0')
        .given("the condition exists")
        .and_given("the user exists", username='alex')
        .upon_receiving("a request").with_request("GET", "/path")
        .will_respond_with(200, body='ok')
    )

    result = pact.construct_pact(pact._interactions[0])
    assert result == {
        'consumer': {'name': 'consumer'},
        'provider': {'name': 'provider'},
        'interactions': [
            {
                'description': 'a request',
                'providerStates': [
                    {'name': 'the condition exists', 'params': {}},
                    {'name': 'the user exists', 'params': {'username': 'alex'}}
                ],
                'request': dict(method='GET', path='/path'),
                'response': dict(status=200, body='ok'),

            }
        ],
        'metadata': dict(pactSpecification=dict(version='3.0.0'))
    }



import io
import logging

import urllib3.connectionpool
import urllib3.poolmanager
from urllib3.response import HTTPResponse

from .pact_request_handler import PactRequestHandler

_providers = {}


log = logging.getLogger(__name__)


class MockPool:
    mocks = {}

    @classmethod
    def add_mock(cls, mock):
        cls.mocks[mock.pact.port] = mock

    @classmethod
    def remove_mock(cls, mock):
        del cls.mocks[mock.pact.port]


class MockConnectionPool(urllib3.connectionpool.HTTPConnectionPool, MockPool):
    def urlopen(self, method, url, body=None, headers=None, *args, **kwargs):
        if self.port not in self.mocks:
            return super().urlopen(method, url, body, headers, *args, **kwargs)
        return self.mocks[self.port](method, url, body=body, headers=headers)


class MonkeyPatcher:
    def __init__(self):
        self.patched = False
        self.handlers = {}

    def add_service(self, handler):
        MockPool.add_mock(handler)
        if not self.patched:
            self.patch()

    def remove_service(self, handler):
        MockPool.remove_mock(handler)
        if not self.handlers:
            self.clear()

    def patch(self):
        urllib3.poolmanager.pool_classes_by_scheme["http"] = MockConnectionPool
        self.patched = True

    def clear(self):
        urllib3.poolmanager.pool_classes_by_scheme["http"] = urllib3.connectionpool.HTTPConnectionPool
        self.patched = False


patcher = MonkeyPatcher()


class MockURLOpenHandler(PactRequestHandler):
    def __init__(self, config):
        self.interactions = []
        super().__init__(config)
        patcher.add_service(self)

    def terminate(self):
        patcher.remove_service(self)

    def setup(self, interactions):
        self.interactions = interactions

    def verify(self):
        pass

    def __call__(self, method, url, redirect=True, headers=None, body=None, **kw):
        self.path = url
        self.headers = headers or {}
        self.body = body
        return self.validate_request(method)

    def get_interaction(self, path):
        try:
            interaction = self.interactions.pop()
        except IndexError:
            raise AssertionError(f'Request at {path} received but no interaction registered') from None
        return interaction

    def handle_failure(self, reason):
        raise AssertionError(reason)

    def handle_success(self, interaction):
        pass

    def respond_for_interaction(self, interaction):
        headers = {}
        if 'headers' in interaction['response']:
            headers.update(interaction['response']['headers'])
        if 'body' in interaction['response']:
            body = self.handle_response_encoding(interaction['response'], headers)
        else:
            body = b''
        return HTTPResponse(body=io.BytesIO(body),
                            status=interaction['response']['status'],
                            preload_content=False,
                            headers=headers)



import pytest
import requests

from pactman import Equals, Consumer, Provider, Like


@pytest.mark.parametrize('object', [None, list(), dict(), 1, 1.0, 'string'])
def test_valid_types(object):
    Equals(object)


@pytest.mark.parametrize('object', [set(), b'bytes'])
def test_invalid_types(object):
    with pytest.raises(AssertionError) as e:
        Equals(object)

    assert 'matcher must be one of ' in str(e.value)


def test_basic_type():
    assert Equals(123).generate_matching_rule_v3() == {'matchers': [{'match': 'equality'}]}


def test_v2_not_allowed():
    with pytest.raises(Equals.NotAllowed):
        Consumer('C').has_pact_with(Provider('P'), version='2.0.0') \
            .given("g").upon_receiving("r").with_request("post", "/foo", body=Equals("bee")) \
            .will_respond_with(200)


def test_mock_usage_pass_validation():
    pact = Consumer('C').has_pact_with(Provider('P'), version='3.0.0') \
        .given("g").upon_receiving("r").with_request("post", "/foo", body=Like({"a": "spam", "b": Equals("bee")})) \
        .will_respond_with(200)

    with pact:
        requests.post(pact.uri + '/foo', json={"a": "ham", "b": "bee"})


def test_mock_usage_fail_validation():
    pact = Consumer('C').has_pact_with(Provider('P'), version='3.0.0') \
        .given("g").upon_receiving("r").with_request("post", "/foo", body=Like({"a": "spam", "b": Equals("bee")})) \
        .will_respond_with(200)

    with pytest.raises(AssertionError), pact:
        requests.post(pact.uri + '/foo', json={"a": "ham", "b": "wasp"})



import json
import os
import tempfile

import requests
from unittest.mock import Mock, mock_open, patch

import pactman.mock.pact
import pytest
import semver
from pactman.mock.consumer import Consumer
from pactman.mock.pact import Pact
from pactman.mock.pact_request_handler import PactRequestHandler, PactVersionConflict, PactInteractionMismatch
from pactman.mock.provider import Provider


@pytest.fixture
def mock_pact(monkeypatch):
    def f(file_write_mode=None, version='2.0.0'):
        monkeypatch.setattr(Pact, "allocate_port", Mock())
        monkeypatch.setattr(os, "remove", Mock())
        monkeypatch.setattr(os.path, "exists", Mock(return_value=True))
        log_dir = "/tmp/a"
        pact_dir = "/tmp/pact"
        return Pact(Consumer("CONSUMER"), Provider("PROVIDER"), log_dir=log_dir, pact_dir=pact_dir, version=version,
                    file_write_mode=file_write_mode)
    return f


@pytest.mark.parametrize("file_write_mode", [None, "overwrite"])
def test_pact_init(monkeypatch, file_write_mode, mock_pact):
    monkeypatch.setattr(pactman.mock.pact, 'ensure_pact_dir', Mock(return_value=True))
    mock_pact = mock_pact(file_write_mode)
    filename = mock_pact.pact_json_filename

    assert(mock_pact.consumer.name == "CONSUMER")
    assert(mock_pact.provider.name == "PROVIDER")
    assert(mock_pact.log_dir == "/tmp/a")
    assert(mock_pact.version == "2.0.0")
    assert(mock_pact.semver == semver.parse("2.0.0"))
    mock_pact.allocate_port.assert_called_once_with()
    assert(mock_pact.BASE_PORT_NUMBER >= 8150)
    pactman.mock.pact.ensure_pact_dir.assert_called_once_with("/tmp/pact")
    if file_write_mode == "overwrite":
        os.path.exists.assert_called_once_with(filename)
        os.remove.assert_called_once_with(filename)
    else:
        os.path.exists.assert_not_called()
        os.remove.assert_not_called()


def test_config_pact_filename(mock_pact):
    mock_pact = mock_pact()
    assert(mock_pact.pact_json_filename == os.path.join(mock_pact.pact_dir, "CONSUMER-PROVIDER-pact.json"))


def test_ensure_pact_dir_when_exists(monkeypatch):
    monkeypatch.setattr(os.path, 'exists', Mock(side_effect=[True]))
    monkeypatch.setattr(os, 'mkdir', Mock())
    pactman.mock.pact.ensure_pact_dir('/tmp/pacts')
    os.mkdir.assert_not_called()


def test_ensure_pact_dir_when_parent_exists(monkeypatch):
    monkeypatch.setattr(os.path, 'exists', Mock(side_effect=[False, True]))
    monkeypatch.setattr(os, 'mkdir', Mock())
    pactman.mock.pact.ensure_pact_dir('/tmp/pacts')
    os.mkdir.assert_called_once_with('/tmp/pacts')


def generate_pact(version):
    return {
        "consumer": {"name": "CONSUMER"},
        "provider": {"name": "PROVIDER"},
        "interactions": [dict(description='spam')],
        "metadata": {
            'pactSpecification': {'version': version}
        }
    }


@pytest.mark.parametrize("version", ["2.0.0", "3.0.0"])
@patch("builtins.open", new_callable=mock_open, read_data="data")
def test_pact_request_handler_write_pact(mock_open, monkeypatch, mock_pact, version):
    monkeypatch.setattr(pactman.mock.pact, 'ensure_pact_dir', Mock(return_value=True))
    mock_pact = mock_pact(version=version)
    mock_pact.semver = semver.parse(version)
    my_pact = PactRequestHandler(mock_pact)
    os.path.exists.return_value = False
    with patch("json.dump", Mock()) as json_mock:
        my_pact.write_pact(dict(description='spam'))
        mock_open.assert_called_once_with(mock_pact.pact_json_filename, "w")
        json_mock.assert_called_once_with(generate_pact(version), mock_open(), indent=2)


@patch("builtins.open", new_callable=mock_open, read_data="data")
def test_versions_are_consistent(mock_open, monkeypatch, mock_pact):
    monkeypatch.setattr(pactman.mock.pact, 'ensure_pact_dir', Mock(return_value=True))
    monkeypatch.setattr(json, 'dump', Mock())
    monkeypatch.setattr(json, 'load', lambda f: generate_pact('2.0.0'))

    # write the v2 pact
    pact = mock_pact()
    pact.semver = semver.parse(pact.version)
    hdlr = PactRequestHandler(pact)
    hdlr.write_pact(dict(description='spam'))

    # try to add the v3 pact
    pact = mock_pact(version='3.0.0')
    pact.semver = semver.parse(pact.version)
    hdlr = PactRequestHandler(pact)
    with pytest.raises(PactVersionConflict):
        hdlr.write_pact(dict(description='spam'))


def test_pacts_written():
    with tempfile.TemporaryDirectory() as d:
        pact = Consumer('C').has_pact_with(Provider('P'), pact_dir=d)
        with pact.given("g").upon_receiving("r").with_request("get", "/foo").will_respond_with(200):
            requests.get(pact.uri + '/foo')

        # force a failure
        with pytest.raises(AssertionError):
            with pact.given("g").upon_receiving("r2").with_request("get", "/bar").will_respond_with(200):
                requests.get(pact.uri + '/foo')

        # make sure mocking still works
        with pact.given("g").upon_receiving("r2").with_request("get", "/bar").will_respond_with(200):
            requests.get(pact.uri + '/bar')

        # ensure two pacts written
        with open(pact.pact_json_filename) as f:
            content = json.load(f)
            assert len(content['interactions']) == 2


def test_detect_mismatch_request_manual_mode():
    with tempfile.TemporaryDirectory() as d:
        pact = Consumer('C').has_pact_with(Provider('P'), pact_dir=d, file_write_mode="merge") \
            .given("g").upon_receiving("r").with_request("get", "/foo").will_respond_with(200)
        with pact:
            requests.get(pact.uri + '/foo')

        # force a failure by specifying the same given/providerState but different request
        pact = Consumer('C').has_pact_with(Provider('P'), pact_dir=d, file_write_mode="merge") \
            .given("g").upon_receiving("r").with_request("get", "/bar").will_respond_with(200)
        with pytest.raises(PactInteractionMismatch):
            with pact:
                requests.get(pact.uri + '/bar')


def test_detect_mismatch_request_retained_relationship():
    with tempfile.TemporaryDirectory() as d:
        pact = Consumer('C').has_pact_with(Provider('P'), pact_dir=d)
        with pact.given("g").upon_receiving("r").with_request("get", "/foo").will_respond_with(200):
            requests.get(pact.uri + '/foo')

        # force a failure by specifying the same given/providerState but different request
        with pytest.raises(PactInteractionMismatch):
            with pact.given("g").upon_receiving("r").with_request("get", "/bar").will_respond_with(200):
                requests.get(pact.uri + '/bar')



import pytest
import requests

from pactman import Consumer, Provider, Like, Includes


def test_valid_types():
    Includes('string', 'sample data')


@pytest.mark.parametrize('object', [None, list(), dict(), set(), 1, 1.0, b'bytes'])
def test_invalid_types(object):
    with pytest.raises(AssertionError) as e:
        Includes(object, 'sample data')

    assert 'matcher must be a string' in str(e.value)


def test_basic_type():
    assert Includes('spam', 'sample data').generate_matching_rule_v3() == \
           {'matchers': [{'match': 'include', 'value': 'spam'}]}


def test_v2_not_allowed():
    with pytest.raises(Includes.NotAllowed):
        Consumer('C').has_pact_with(Provider('P'), version='2.0.0') \
            .given("g").upon_receiving("r").with_request("post", "/foo", body=Includes("bee", 'been')) \
            .will_respond_with(200)


def test_mock_usage_pass_validation():
    pact = Consumer('C').has_pact_with(Provider('P'), version='3.0.0') \
        .given("g").upon_receiving("r").with_request("post", "/foo", body=Like({"a": "spam",
                                                                                "b": Includes("bee", 'been')})) \
        .will_respond_with(200)

    with pact:
        requests.post(pact.uri + '/foo', json={"a": "ham", "b": "has bee in it"})


def test_mock_usage_fail_validation():
    pact = Consumer('C').has_pact_with(Provider('P'), version='3.0.0') \
        .given("g").upon_receiving("r").with_request("post", "/foo", body=Like({"a": "spam",
                                                                                "b": Includes("bee", 'been')})) \
        .will_respond_with(200)

    with pytest.raises(AssertionError), pact:
        requests.post(pact.uri + '/foo', json={"a": "ham", "b": "wasp"})



"""Classes for defining request and response data that is variable."""


class Matcher(object):
    """Base class for defining complex contract expectations."""

    def ruby_protocol(self):  # pragma: no cover
        """
        Serialise this Matcher for the Ruby mocking server.

        :rtype: any
        """
        raise NotImplementedError

    def generate_matching_rule_v3(self):  # pragma: no cover
        raise NotImplementedError


class EachLike(Matcher):
    """
    Expect the data to be a list of similar objects.

    Example:

    >>> from pactman import Consumer, Provider
    >>> pact = Consumer('consumer').has_pact_with(Provider('provider'))
    >>> (pact.given('there are three comments')
    ...  .upon_receiving('a request for the most recent 2 comments')
    ...  .with_request('get', '/comment', query={'limit': 2})
    ...  .will_respond_with(200, body={
    ...    'comments': EachLike(
    ...        {'name': Like('bob'), 'text': Like('Hello!')},
    ...        minimum=2)
    ...  }))

    Would expect the response to be a JSON object, with a comments list. In
    that list should be at least 2 items, and each item should be a `dict`
    with the keys `name` and `text`,
    """

    def __init__(self, matcher, minimum=1):
        """
        Create a new EachLike.

        :param matcher: The expected value that each item in a list should
            look like, this can be other matchers.
        :type matcher: None, list, dict, int, float, str, unicode, Matcher
        :param minimum: The minimum number of items expected.
            Must be greater than or equal to 1.
        :type minimum: int
        """
        self.matcher = matcher
        if minimum < 1:
            raise AssertionError('Minimum must be greater than or equal to 1')
        self.minimum = minimum

    def ruby_protocol(self):
        """
        Serialise this EachLike for the Ruby mocking server.

        :return: A dict containing the information about the contents of the
            list and the provided minimum number of items for that list.
        :rtype: dict
        """
        return {
            'json_class': 'Pact::ArrayLike',
            'contents': generate_ruby_protocol(self.matcher),
            'min': self.minimum}

    def generate_matching_rule_v3(self):
        return {'matchers': [{'match': 'type', 'min': self.minimum}]}


class Like(Matcher):
    """
    Expect the type of the value to be the same as matcher.

    Example:

    >>> from pactman import Consumer, Provider
    >>> pact = Consumer('consumer').has_pact_with(Provider('provider'))
    >>> (pact
    ...  .given('there is a random number generator')
    ...  .upon_receiving('a request for a random number')
    ...  .with_request('get', '/generate-number')
    ...  .will_respond_with(200, body={
    ...    'number': Like(1111222233334444)
    ...  }))

    Would expect the response body to be a JSON object, containing the key
    `number`, which would contain an integer. When the consumer runs this
    contract, the value `1111222233334444` will be returned by the mock
    service, instead of a randomly generated value.
    """

    def __init__(self, matcher):
        """
        Create a new Like.

        :param matcher: The object that should be expected. The mock
            will return this value. When verified against the provider, the
            type of this value will be asserted, while the value will be
            ignored.
        :type matcher: None, list, dict, int, float, str, Matcher
        """
        valid_types = (
            type(None), list, dict, int, float, str, Matcher)

        assert isinstance(matcher, valid_types), (
            "matcher must be one of '{}', got '{}'".format(
                valid_types, type(matcher)))

        self.matcher = matcher

    def ruby_protocol(self):
        """
        Serialise this Like for the Ruby mocking server.

        :return: A dict containing the information about what the contents of
            the request/response should be.
        :rtype: dict
        """
        return {
            'json_class': 'Pact::SomethingLike',
            'contents': generate_ruby_protocol(self.matcher)}

    def generate_matching_rule_v3(self):
        return {'matchers': [{'match': 'type'}]}


# Remove SomethingLike in major version 1.0.0
SomethingLike = Like


class Term(Matcher):
    """
    Expect the response to match a specified regular expression.

    Example:

    >>> from pactman import Consumer, Provider
    >>> pact = Consumer('consumer').has_pact_with(Provider('provider'))
    >>> (pact.given('the current user is logged in as `tester`')
    ...  .upon_receiving('a request for the user profile')
    ...  .with_request('get', '/profile')
    ...  .will_respond_with(200, body={
    ...    'name': 'tester',
    ...    'theme': Term('light|dark|legacy', 'dark')
    ...  }))

    Would expect the response body to be a JSON object, containing the key
    `name`, which will contain the value `tester`, and `theme` which must be
    one of the values: light, dark, or legacy. When the consumer runs this
    contract, the value `dark` will be returned by the mock.
    """

    def __init__(self, matcher, generate):
        """
        Create a new Term.

        :param matcher: A regular expression to find.
        :type matcher: basestring
        :param generate: A value to be returned by the mock when
            generating the response to the consumer.
        :type generate: basestring
        """
        self.matcher = matcher
        self.generate = generate

    def ruby_protocol(self):
        """
        Serialise this Term for the Ruby mocking server.

        :return: A dict containing the information about what the contents of
            the request/response should be, and what should match for the requests.
        :rtype: dict
        """
        return {
            'json_class': 'Pact::Term',
            'data': {
                'generate': self.generate,
                'matcher': {
                    'json_class': 'Regexp',
                    'o': 0,
                    's': self.matcher}}}

    def generate_matching_rule_v3(self):
        return {'matchers': [{'match': 'regex', 'regex': self.matcher}]}


class Equals(Matcher):
    """
    Expect the value to be the same as matcher.

    Example:

    >>> from pactman import Consumer, Provider
    >>> pact = Consumer('consumer').has_pact_with(Provider('provider'))
    >>> (pact
    ...  .given('there is a random number generator')
    ...  .upon_receiving('a request for a random number')
    ...  .with_request('get', '/generate-number')
    ...  .will_respond_with(200, body={
    ...    'number': Equals(1111222233334444)
    ...  }))

    Would expect the response body to be a JSON object, containing the key
    `number`, which would contain the value `1111222233334444`.
    When the consumer runs this contract, the value `1111222233334444`
    will be returned by the mock, instead of a randomly generated value.
    """

    class NotAllowed(TypeError):
        pass

    def __init__(self, matcher):
        """
        Create a new Equals.

        :param matcher: The object that should be expected. The mock
            will return this value. When verified against the provider, the
            value will be asserted.
        :type matcher: None, list, dict, int, float, str
        """
        valid_types = (type(None), list, dict, int, float, str)
        assert isinstance(matcher, valid_types), f"matcher must be one of '{valid_types}', got '{type(matcher)}'"

        self.matcher = matcher

    def generate_matching_rule_v3(self):
        return {'matchers': [{'match': 'equality'}]}


class Includes(Matcher):
    """
    Expect the string value to contain the matcher.

    Example:

    >>> from pactman import Consumer, Provider
    >>> pact = Consumer('consumer').has_pact_with(Provider('provider'))
    >>> (pact
    ...  .given('there is a random number generator')
    ...  .upon_receiving('a request for a random number')
    ...  .with_request('get', '/generate-number')
    ...  .will_respond_with(200, body={
    ...    'content': Includes('spam', 'Some example spamming content')
    ...  }))

    Would expect the response body to be a JSON object, containing the key
    `content`, which be a string containing `'spam'`.
    When the consumer runs this contract, the value `'Some example spamming content'`
    will be returned by the mock.
    """

    class NotAllowed(TypeError):
        pass

    def __init__(self, matcher, generate):
        """
        Create a new Includes.

        :param matcher: The substring that should be expected. When verified against the
            provider, the value will be asserted.
        :type matcher: string
        :param generate: The mock will return this value.
        :type generate: string
        """
        assert isinstance(matcher, str), f"matcher must be a string, got '{type(matcher)}'"

        self.matcher = matcher
        self.generate = generate

    def generate_matching_rule_v3(self):
        return {'matchers': [{'match': 'include', 'value': self.matcher}]}


def generate_ruby_protocol(term):
    """
    Parse the provided term into the JSON for the Ruby mock server.

    :param term: The term to be parsed.
    :type term: None, list, dict, int, float, str, unicode, Matcher
    :return: The JSON representation for this term.
    :rtype: dict, list, str
    """
    if term is None:
        return term
    elif isinstance(term, (str, int, float)):
        return term
    elif isinstance(term, dict):
        return {k: generate_ruby_protocol(v) for k, v in term.items()}
    elif isinstance(term, list):
        return [generate_ruby_protocol(t) for i, t in enumerate(term)]
    elif issubclass(term.__class__, (Matcher,)):
        return term.ruby_protocol()
    else:
        raise ValueError('Unknown type: %s' % type(term))


# For backwards compatiblity with test code that uses pact-python to declare pacts,
# allow the various classes from that package to also be used to define rules
try:
    import pact as pact_python
    LIKE_CLASSES = (Like, pact_python.Like)
    EACHLIKE_CLASSES = (EachLike, pact_python.EachLike)
    TERM_CLASSES = (Term, pact_python.Term)
except ImportError:
    pact_python = None
    LIKE_CLASSES = (Like,)
    EACHLIKE_CLASSES = (EachLike,)
    TERM_CLASSES = (Term,)


# this function is long and complex (C901) because it has to handle the pact-python
# types :-(
def get_generated_values(input):  # noqa: C901
    """
    Resolve (nested) Matchers to their generated values for assertion.

    :param input: The input to be resolved to its generated values.
    :type input: None, list, dict, int, float, bool, str, unicode, Matcher
    :return: The input resolved to its generated value(s)
    :rtype: None, list, dict, int, float, bool, str, unicode, Matcher
    """
    if input is None:
        return input
    if isinstance(input, (str, int, float, bool)):
        return input
    if isinstance(input, dict):
        return {k: get_generated_values(v) for k, v in input.items()}
    if isinstance(input, list):
        return [get_generated_values(t) for i, t in enumerate(input)]
    elif isinstance(input, LIKE_CLASSES):
        return get_generated_values(input.matcher)
    elif isinstance(input, EACHLIKE_CLASSES):
        return [get_generated_values(input.matcher)] * input.minimum
    elif isinstance(input, Term):
        return input.generate
    elif pact_python is not None and isinstance(input, pact_python.Term):
        return input._generate
    elif isinstance(input, Equals):
        return get_generated_values(input.matcher)
    elif isinstance(input, Includes):
        return input.generate
    else:
        raise ValueError('Unknown type: %s' % type(input))


# this function is long and complex (C901) because it has to handle the pact-python
# types :-(
def get_matching_rules_v2(input, path):  # noqa: C901
    '''Turn a matcher into the matchingRules structure for pact JSON.

    This is done recursively, adding new paths as new matching rules
    are encountered.
    '''
    if input is None or isinstance(input, (str, int, float, bool)):
        return {}
    if isinstance(input, dict):
        rules = {}
        for k, v in input.items():
            sub_path = path + '.' + k
            rules.update(get_matching_rules_v2(v, sub_path))
        return rules
    if isinstance(input, list):
        rules = {}
        for i, v in enumerate(input):
            sub_path = path + '[*]'
            rules.update(get_matching_rules_v2(v, sub_path))
        return rules
    if isinstance(input, LIKE_CLASSES):
        rules = {path: {'match': 'type'}}
        rules.update(get_matching_rules_v2(input.matcher, path))
        return rules
    if isinstance(input, EACHLIKE_CLASSES):
        rules = {path: {'match': 'type', 'min': input.minimum}}
        rules.update(get_matching_rules_v2(input.matcher, path))
        return rules
    if isinstance(input, TERM_CLASSES):
        return {path: {'regex': input.matcher}}
    if isinstance(input, Equals):
        raise Equals.NotAllowed(f'Equals() cannot be used in pact version 2')
    if isinstance(input, Includes):
        raise Includes.NotAllowed(f'Includes() cannot be used in pact version 2')

    raise ValueError('Unknown type: %s' % type(input))


class MatchingRuleV3(dict):
    def generate(self, input, path):
        if self.handle_basic_types(input, path):
            return
        if self.handle_pactman_types(input, path):
            return
        if self.handle_pact_python_types(input, path):
            return
        raise ValueError('Unknown type: %s' % type(input))

    def handle_basic_types(self, input, path):
        if input is None or isinstance(input, (str, int, float, bool)):
            return True
        if isinstance(input, dict):
            for k, v in input.items():
                self.generate(v, path + '.' + k)
            return True
        if isinstance(input, list):
            for v in input:
                self.generate(v, path + '[*]')
            return True
        return False

    def handle_pactman_types(self, input, path):
        if not hasattr(input, 'generate_matching_rule_v3'):
            return False
        self[path] = input.generate_matching_rule_v3()
        if isinstance(input.matcher, (list, dict)):
            self.handle_basic_types(input.matcher, path)
        return True

    def handle_pact_python_types(self, input, path):
        if pact_python is None:
            return False

        if isinstance(input, pact_python.Like):
            self[path] = {'matchers': [{'match': 'type'}]}
            self.generate(input.matcher, path)
        elif isinstance(input, pact_python.EachLike):
            self[path] = {'matchers': [{'match': 'type', 'min': input.minimum}]}
            self.generate(input.matcher, path)
        elif isinstance(input, pact_python.Term):
            self[path] = {'matchers': [{'match': 'regex', 'regex': input.matcher}]}
        else:
            return False

        return True


def get_matching_rules_v3(input, path):
    '''Turn a matcher into the matchingRules structure for pact JSON.

    This is done recursively, adding new paths as new matching rules
    are encountered.
    '''
    rules = MatchingRuleV3()
    rules.generate(input, path)
    return rules



"""Python methods for interactive with a Pact Mock Service."""
from .mock.consumer import Consumer
from .mock.matchers import EachLike, Equals, Includes, Like, SomethingLike, Term
from .mock.pact import Pact
from .mock.provider import Provider

__all__ = ('Consumer', 'EachLike', 'Equals', 'Includes', 'Like', 'Pact', 'Provider',
           'SomethingLike', 'Term')



from pactman import Consumer, Provider


def test_full_payload_v2():
    pact = Consumer('consumer').has_pact_with(Provider('provider'), version='2.0.0')
    (pact
        .given('UserA exists and is not an administrator')
        .upon_receiving('a request for UserA')
        .with_request('get', '/users/UserA', headers={'Accept': 'application/json'}, query='term=test')
        .will_respond_with(200, body={'username': 'UserA'}, headers={'Content-Type': 'application/json'}))
    result = pact.construct_pact(pact._interactions[0])
    assert result == {
        'consumer': {'name': 'consumer'},
        'provider': {'name': 'provider'},
        'interactions': [
            {
                'description': 'a request for UserA',
                'providerState': 'UserA exists and is not an administrator',
                'request': dict(method='get', path='/users/UserA', headers={'Accept': 'application/json'},
                                query='term=test'),
                'response': dict(status=200, body={'username': 'UserA'},
                                 headers={'Content-Type': 'application/json'})
            }
        ],
        'metadata': dict(pactSpecification=dict(version='2.0.0'))
    }


def test_full_payload_v3():
    pact = Consumer('consumer').has_pact_with(Provider('provider'), version='3.0.0')
    (pact
     .given([{"name": "User exists and is not an administrator", "params": {"username": "UserA"}}])
     .upon_receiving('a request for UserA')
     .with_request('get', '/users/UserA', headers={'Accept': 'application/json'}, query=dict(term=['test']))
     .will_respond_with(200, body={'username': 'UserA'}, headers={'Content-Type': 'application/json'}))
    result = pact.construct_pact(pact._interactions[0])
    assert result == {
        'consumer': {'name': 'consumer'},
        'provider': {'name': 'provider'},
        'interactions': [
            {
                'description': 'a request for UserA',
                'providerStates': [{
                    "name": "User exists and is not an administrator",
                    "params": {"username": "UserA"}
                }],
                'request': dict(method='get', path='/users/UserA', headers={'Accept': 'application/json'},
                                query=dict(term=['test'])),
                'response': dict(status=200, body={'username': 'UserA'},
                                 headers={'Content-Type': 'application/json'})
            }
        ],
        'metadata': dict(pactSpecification=dict(version='3.0.0'))
    }



from unittest.mock import Mock

from pactman.verifier import result


def test_logged_start_result(monkeypatch):
    monkeypatch.setattr(result, 'log', Mock())
    r = result.LoggedResult()
    r.start(Mock())
    result.log.info.assert_called_once()


def test_logged_fail_result(monkeypatch):
    monkeypatch.setattr(result, 'log', Mock())
    r = result.LoggedResult()
    r.fail('message')
    result.log.warning.assert_called_once_with(' message')


def test_logged_fail_result_path(monkeypatch):
    monkeypatch.setattr(result, 'log', Mock())
    r = result.LoggedResult()
    r.fail('message', ['a', 0])
    result.log.warning.assert_called_once_with(' message')
